import tensorflow as tf
import keras


class AngularMinPooling(keras.layers.Layer):
    """The implementation for angular max-pooling"""

    @tf.function
    def call(self, inputs):
        """Min-pools over the results of a intrinsic surface convolution.

        Parameters
        ----------
        inputs: tensorflow.Tensor
            The result tensor of an intrinsic surface convolution.
            It has a size of: (n_vertices, n_rotations, feature_dim), where 'n_vertices' references to the total amount
            of vertices in the triangle mesh, 'n_rotations' to the amount of rotations considered during the intrinsic
            surface convolution and 'feature_dim' to the feature dimensionality.

        Returns
        -------
        tensorflow.Tensor:
            A two-dimensional tensor of size (n_vertices, feature_dim), that contains a convolution results for each
            vertex. Thereby, the convolution result has the smallest Euclidean norm among the convolution results for
            all rotations.
        """
        minimal_response = tf.norm(inputs, ord="euclidean", axis=-1)
        minimal_response = tf.cast(tf.argmin(minimal_response, axis=1), dtype=tf.int32)
        return tf.gather_nd(
            indices=tf.stack([tf.range(tf.shape(inputs)[0]), minimal_response], axis=-1),
            params=inputs
        )



from geoconv_examples.mpi_faust.pytorch.faust_data_set import FaustDataset
from geoconv_examples.mpi_faust.pytorch.model import Imcnn
from geoconv_examples.mpi_faust.data.preprocess_faust import preprocess_faust
from geoconv.utils.measures import princeton_benchmark

from pathlib import Path
from torch import nn

import torch
import numpy as np
import sys
import json
import os


def train_model(reference_mesh_path,
                signal_dim,
                preprocessed_data,
                n_radial=5,
                n_angular=8,
                registration_path="",
                compute_shot=True,
                geodesic_diameters_path="",
                precomputed_gpc_radius=0.037,
                template_radius=0.027744965069279016,
                logging_dir="./imcnn_training_logs",
                processes=1,
                init_lr=0.00165,
                weight_decay=0.005,
                layer_conf=None,
                model="dirac",
                add_noise=False,
                reference_mesh_diameter=2.2093810817030244,
                early_stop=20):
    """Trains one singular IMCNN

    Parameters
    ----------
    reference_mesh_path: str
        The path to the reference mesh file.
    signal_dim: int
        The dimensionality of the mesh signal
    preprocessed_data: str
        The path to the pre-processed data. If you have not pre-processed your data so far and saved it under the given
        path, this script will execute pre-processing for you. For this to work, you need to pass the arguments which
        are annotated with '[REQUIRED FOR PRE-PROCESSING]'. If pre-processing is not required, you can ignore those
        arguments.
    reference_mesh_diameter: float
        [REQUIRED FOR BENCHMARKING] The geodesic diameter of the reference mesh. Defaults to the diameter of the first
        registration mesh (tr_reg_000.ply) in the training set of the FAUST data set. If other reference mesh is
        selected, adjust this parameter accordingly!
    n_radial: int
        [REQUIRED FOR PRE-PROCESSING] The amount of radial coordinates for the template.
    n_angular: int
        [REQUIRED FOR PRE-PROCESSING] The amount of angular coordinates for the template.
    registration_path: str
        [REQUIRED FOR PRE-PROCESSING] The path of the training-registration files in the FAUST data set.
    compute_shot: bool
        [REQUIRED FOR PRE-PROCESSING] Whether to compute SHOT-descriptors during preprocessing as the mesh signal
    geodesic_diameters_path: str
        [REQUIRED FOR PRE-PROCESSING] The path to pre-computed geodesic diameters for the FAUST-registration meshes.
    precomputed_gpc_radius: float
        [REQUIRED FOR PRE-PROCESSING] The GPC-system radius to use for GPC-system computation. If not provided, the
        script will calculate it.
    template_radius: float
        [OPTIONAL] The template radius of the ISC-layer (the one used during preprocessing, defaults to radius for FAUST
        data set).
    logging_dir: str
        [OPTIONAL] The path to the folder where logs will be stored
    processes: int
        [OPTIONAL] The amount of concurrent processes. Affects preprocessing and Princeton benchmark.
    init_lr: float
        [OPTIONAL] Initial learning rate.
    weight_decay: float
        [OPTIONAL] Weight decay.
    layer_conf: list
        [OPTIONAL] List of tuples: The first entry references the output dimensions of the i-th ISC-layer, The second
        entry references of skips between each rotation while computing the convolution (rotation delta).
    model: str
        [OPTIONAL] Which model variant (['dirac', 'geodesic', 'zero']) shall be tuned.
    add_noise: bool
        [OPTIONAL] Adds Gaussian noise to the mesh data.
    early_stop: int
        [OPTIONAL] The amount of epochs for early stopping.
    """
    # Create logging dir
    if not os.path.exists(logging_dir):
        os.makedirs(logging_dir)

    # Load data
    preprocess_zip = f"{preprocessed_data}.zip"
    if not Path(preprocess_zip).is_file():
        template_radius = preprocess_faust(
            n_radial=n_radial,
            n_angular=n_angular,
            target_dir=preprocess_zip[:-4],
            registration_path=registration_path,
            shot=compute_shot,
            geodesic_diameters_path=geodesic_diameters_path,
            precomputed_gpc_radius=precomputed_gpc_radius,
            processes=processes,
            add_noise=add_noise
        )
    else:
        print(f"Found preprocess-results: '{preprocess_zip}'. Skipping preprocessing.")

    # Check for GPU
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    seeds = [10, 20, 30, 40, 50]
    for exp_number in range(len(seeds)):
        # Set seeds
        torch.manual_seed(seeds[exp_number])
        np.random.seed(seeds[exp_number])
        sys.stdout.write(f"\n### Experiment no. {exp_number}")

        # Define model
        imcnn = Imcnn(
            signal_dim=signal_dim,
            kernel_size=(n_radial, n_angular),
            template_radius=template_radius,
            adapt_data=FaustDataset(preprocess_zip, set_type=0, only_signal=True, device=device),
            layer_conf=layer_conf,
            variant=model
        )
        imcnn.to(device)
        imcnn = torch.compile(imcnn)

        # Define loss, optimizer and scheduler
        loss_fn = nn.CrossEntropyLoss()
        opt = torch.optim.AdamW(
            params=imcnn.parameters(), lr=init_lr, weight_decay=weight_decay, eps=1e-07
        )
        opt.param_groups[0]["initial_lr"] = init_lr

        # Fit model
        training_history = {}
        best_loss = np.inf
        stale_counter = 0
        for epoch in range(200):
            sys.stdout.write("\n")  # pretty printing

            # Training
            train_data = FaustDataset(preprocess_zip, set_type=0, device=device)
            train_dict = imcnn.train_loop(
                train_data,
                loss_fn,
                opt,
                decay_rate=0.99,
                decay_steps=500,
                verbose=True,
                epoch=epoch,
                prev_steps=epoch * 70
            )

            # Validation
            val_data = FaustDataset(preprocess_zip, set_type=1, device=device)
            val_dict = imcnn.validation_loop(val_data, loss_fn, verbose=True)
            val_loss = val_dict["val_epoch_loss"].item()

            # Remember epoch statistics
            training_history[f"epoch_{epoch}"] = {
                "epoch": epoch,
                "loss": train_dict["epoch_loss"].item(),
                "Accuracy": train_dict["epoch_accuracy"].item(),
                "Validation Loss": val_loss,
                "Validation Accuracy": val_dict["val_epoch_accuracy"].item()
            }

            # Log best model
            if val_loss < best_loss:
                best_loss = val_loss
                imcnn_path = f"{logging_dir}/imcnn_exp_{exp_number}_epoch_{epoch}"
                torch.save(imcnn.state_dict(), imcnn_path)
                stale_counter = 0
            else:
                stale_counter += 1

            # Early stopping
            if stale_counter >= early_stop:
                sys.stdout.write("\nEarly stopping.")
                break

        print("\n")  # pretty printing

        # Log training statistics
        with open(f"{logging_dir}/training_history_{exp_number}.json", "w") as file:
            json.dump(training_history, file, indent=4)

        test_dataset = FaustDataset(preprocess_zip, set_type=2, device=device)
        with torch.no_grad():
            princeton_benchmark(
                imcnn=imcnn,
                test_dataset=test_dataset,
                ref_mesh_path=reference_mesh_path,
                file_name=f"{logging_dir}/model_benchmark_{exp_number}",
                processes=processes,
                geodesic_diameter=reference_mesh_diameter,
                pytorch_model=True
            )



import tensorflow as tf
import keras


class ISCResidual(keras.Model):
    def __init__(self,
                 first_isc,
                 second_isc,
                 pool,
                 template_radius,
                 layer_conf=None,
                 splits=1,
                 activation="relu",
                 fit_dim=False):
        super().__init__()

        if layer_conf is None:
            self.output_dims = [64, 64]
            self.rotation_deltas = [1, 1]
        else:
            self.output_dims, self.rotation_deltas = list(zip(*layer_conf))

        self.pool = pool()
        self.first_isc = first_isc(
            amt_templates=self.output_dims[0],
            template_radius=template_radius,
            activation=activation,
            name="residual_1",
            splits=splits,
            rotation_delta=self.rotation_deltas[0]
        )
        self.second_isc = second_isc(
            amt_templates=self.output_dims[1],
            template_radius=template_radius,
            activation="linear",
            name="residual_2",
            splits=splits,
            rotation_delta=self.rotation_deltas[1]
        )

        self.fit_dim = fit_dim
        if self.fit_dim:
            self.third_isc = first_isc(
                amt_templates=self.output_dims[0],
                template_radius=template_radius,
                activation=activation,
                name="residual_1",
                splits=splits,
                rotation_delta=self.rotation_deltas[0]
            )

        self.add = keras.layers.Add()
        self.activation = keras.layers.Activation(activation)

    def call(self, inputs, *args, **kwargs):
        entry_signal, bc = inputs

        signal = self.first_isc([entry_signal, bc])
        signal = self.pool(signal)

        signal = self.second_isc([signal, bc])
        signal = self.pool(signal)

        if self.fit_dim:
            entry_signal = self.third_isc([entry_signal, bc])
            entry_signal = self.pool(entry_signal)

        return self.activation(self.add([entry_signal, signal]))



from geoconv.preprocessing.gpc_system_group import GPCSystemGroup
from geoconv.utils.visualization import draw_gpc_triangles, draw_gpc_on_mesh, draw_barycentric_coordinates
from geoconv.utils.misc import reconstruct_template, find_largest_one_hop_dist, normalize_mesh
from geoconv.preprocessing.barycentric_coordinates import compute_barycentric_coordinates, create_template_matrix

import open3d as o3d
import trimesh
import os
import numpy as np


def load_bunny(path, target_triangles_amount=6000):
    """Loads and simplifies the Stanford bunny

    Download the Stanford bunny from here:
    https://github.com/alecjacobson/common-3d-test-models/blob/master/data/stanford-bunny.zip

    Unzip the .zip-file and move the 'bun_zipper.ply'-file into the folder where this demo-file
    is saved.

    Parameters
    ----------
    target_triangles_amount: int
        The amount of target triangles of the stanford bunny triangle mesh. Watch out, the target
        amount of triangles is not guaranteed!
    path: str
        If the path to the Stanford bunny .ply-file differs from 'PATH_TO_STANFORD_BUNNY' you can
        set it correctly with the 'path'-argument.

    Returns
    -------
    trimesh.Trimesh:
        The Stanford bunny triangle mesh.
    """

    b = o3d.io.read_triangle_mesh(path)
    b = b.simplify_quadric_decimation(target_number_of_triangles=target_triangles_amount)
    b = trimesh.Trimesh(vertices=b.vertices, faces=b.triangles)

    return b


def preprocess_demo(path_to_stanford_bunny="bun_zipper.ply",
                    n_radial=3,
                    n_angular=8,
                    processes=1):
    """Demonstrates and visualizes GeoConv's preprocessing at the hand of the stanford bunny.

    Download the Stanford bunny from here:
    https://github.com/alecjacobson/common-3d-test-models/blob/master/data/stanford-bunny.zip

    Unzip the .zip-file and move the 'bun_zipper.ply'-file into the folder where this demo-file
    is saved.

    Parameters
    ----------
    path_to_stanford_bunny: str
        The path to the 'bun_zipper.ply'-file containing the stanford bunny.
    n_radial: int
        The amount of radial coordinates for the template in your geodesic convolution.
    n_angular: int
        The amount of angular coordinates for the template in your geodesic convolution.
    processes: int
        The amount of concurrent processes that shall compute the GPC-systems.
    """
    target_dir = "./preprocess_demo_plots"
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    # Load the stanford bunny in '.ply'-format as a trimesh-object.
    bunny = load_bunny(path_to_stanford_bunny)

    # Find the smallest distance from a center vertex in the bunny mesh to one of its one-hop neighbors.
    bunny, _ = normalize_mesh(bunny, geodesic_diameter=0.25270776231631265)
    u_max = find_largest_one_hop_dist(bunny)

    # Set the maximal radial distance for the GPC-systems
    u_max = u_max + u_max * .1

    # Select the template radius to be 3/4-th of the GPC-systems max-radius to increase the likelihood of
    # the template vertices to fall into the GPC-system and not exceed its bounds.
    template_radius = u_max * 0.75
    print(f"GPC-system max.-radius: {u_max} | Template max.-radius: {template_radius}")

    # Compute and store the GPC-systems for the bunny mesh.
    gpc_systems = GPCSystemGroup(bunny, processes=processes)
    gpc_systems.compute(u_max=u_max)

    # Compute the barycentric coordinates for the template in the computed GPC-systems.
    bc = compute_barycentric_coordinates(
        gpc_systems, n_radial=n_radial, n_angular=n_angular, radius=template_radius
    )
    np.save(f"{os.path.dirname(path_to_stanford_bunny)}/bunny_barycentric_coordinates.npy", bc)

    ####################################################################
    # Visualization of the GPC-systems and the barycentric coordinates
    ####################################################################
    for gpc_system_idx in range(3):  # select which GPC-systems you want to visualize by altering the list

        # Original template
        template_matrix = create_template_matrix(
            n_radial=n_radial, n_angular=n_angular, radius=template_radius, in_cart=True
        )
        # Create template with barycentric coordinates
        reconstructed_template = reconstruct_template(
            gpc_systems.object_mesh_gpc_systems[gpc_system_idx].get_gpc_system(), bc[gpc_system_idx]
        )

        for radial_coordinate in range(bc.shape[1]):
            for angular_coordinate in range(bc.shape[2]):
                print(
                    f"Location: {(gpc_system_idx, radial_coordinate, angular_coordinate)}: "
                    f"Vertices: {bc[gpc_system_idx, radial_coordinate, angular_coordinate, :, 0]} "
                    f"B.c: {bc[gpc_system_idx, radial_coordinate, angular_coordinate, :, 1]} "
                    f"-> Caused: {reconstructed_template[radial_coordinate, angular_coordinate]}"
                )
        print("==========================================================================")

        # Draw GPC-system on the mesh (center_vertex, radial_coordinates, angular_coordinates, object_mesh)
        draw_gpc_on_mesh(
            gpc_system_idx,
            gpc_systems.object_mesh_gpc_systems[gpc_system_idx].radial_coordinates,
            gpc_systems.object_mesh_gpc_systems[gpc_system_idx].angular_coordinates,
            bunny
        )

        # Original/Set template vertices
        draw_gpc_triangles(
            gpc_systems.object_mesh_gpc_systems[gpc_system_idx],
            template_matrix=template_matrix,
            plot=True,
            title="GPC-system",
            save_name=f"{target_dir}/gpc_system_{gpc_system_idx}"
        )

        # Draw triangles with stored GPC-coordinates and place reconstructed template vertices with the help of the
        # computed barycentric coordinates
        draw_barycentric_coordinates(
            gpc_systems.object_mesh_gpc_systems[gpc_system_idx],
            bc[gpc_system_idx],
            save_name=f"{target_dir}/gpc_system_{gpc_system_idx}"
        )



from torch import nn

import torch


class AngularMaxPooling(nn.Module):
    """The implementation for angular max-pooling"""

    def forward(self, inputs):
        """Min-pools over the results of a intrinsic surface convolution.

        Parameters
        ----------
        inputs: torch.Tensor
            The result tensor of an intrinsic surface convolution.
            It has a size of: (n_vertices, n_rotations, feature_dim), where 'n_vertices' references to the total amount
            of vertices in the triangle mesh, 'n_rotations' to the amount of rotations considered during the intrinsic
            surface convolution and 'feature_dim' to the feature dimensionality.

        Returns
        -------
        torch.Tensor:
            A two-dimensional tensor of size (n_vertices, feature_dim), that contains a convolution results for each
            vertex. Thereby, the convolution result has the smallest Euclidean norm among the convolution results for
            all rotations.
        """
        maximal_response = torch.linalg.vector_norm(inputs, ord=2, dim=-1)
        maximal_response = torch.argmin(maximal_response, dim=1).int()
        return inputs[torch.arange(0, inputs.shape[0]), maximal_response]



from geoconv.pytorch.layers.angular_max_pooling import AngularMaxPooling
from geoconv.pytorch.layers.conv_dirac import ConvDirac
from geoconv.pytorch.layers.conv_geodesic import ConvGeodesic
from geoconv.pytorch.layers.conv_zero import ConvZero

from torch import nn
from torcheval.metrics.functional import multiclass_accuracy

import torch
import sys


def custom_exp_scheduler(opt, step, decay_rate=0.95, decay_steps=500):
    """Smooth exponential scheduler"""
    opt.param_groups[0]["lr"] = opt.param_groups[0]["initial_lr"] * decay_rate ** (step / decay_steps)


def print_mem():
    mem = torch.cuda.memory_allocated()
    max_mem = torch.cuda.max_memory_allocated()
    return f"{mem / 1024 ** 2:.3f} MB / Max memory: {max_mem / 1024 ** 2:.3f} MB"


class Normalization(nn.Module):

    def __init__(self, dataset):
        super().__init__()
        self.mean, self.var = 0, 0
        n_samples = 0

        for s in dataset:
            n_samples += s.shape[0]
            s = torch.sum(s, dim=-2)
            self.mean += s
        self.mean = self.mean / n_samples

        dataset.reset()
        for s in dataset:
            self.var += torch.sum((s - self.mean) ** 2, dim=-2)
        self.var = self.var / (n_samples - 1)

    def forward(self, inputs):
        return (inputs - self.mean) / self.var


class Imcnn(nn.Module):
    def __init__(self,
                 signal_dim,
                 kernel_size,
                 template_radius,
                 adapt_data,
                 layer_conf=None,
                 variant="dirac",
                 segmentation_classes=-1):
        super().__init__()
        self.signal_dim = signal_dim
        self.kernel_size = kernel_size
        self.template_radius = template_radius

        if variant == "dirac":
            self.layer_type = ConvDirac
        elif variant == "geodesic":
            self.layer_type = ConvGeodesic
        elif variant == "zero":
            self.layer_type = ConvZero
        else:
            raise RuntimeError("Select a layer type from: ['dirac', 'geodesic', 'zero']")

        if layer_conf is None:
            self.output_dims = [96, 256, 384, 384]
            self.rotation_deltas = [1 for _ in range(len(self.output_dims))]
        else:
            self.output_dims, self.rotation_deltas = list(zip(*layer_conf))
        self.downsize_dim = 64
        self.input_dims = [self.downsize_dim]
        self.input_dims.extend(self.output_dims)

        #################
        # Handling Input
        #################
        self.normalize = Normalization(adapt_data)
        self.downsize_dense = nn.Linear(in_features=signal_dim, out_features=self.downsize_dim)
        self.downsize_activation = nn.ReLU()
        self.downsize_bn = nn.BatchNorm1d(num_features=self.downsize_dim)

        #############
        # ISC Layers
        #############
        self.do_layers = nn.ModuleList()
        self.isc_layers = nn.ModuleList()
        self.bn_layers = nn.ModuleList()
        self.amp_layers = nn.ModuleList()
        for idx in range(len(self.output_dims)):
            self.do_layers.append(nn.Dropout(p=0.2))
            self.isc_layers.append(
                self.layer_type(
                    input_shape=[(None, self.input_dims[idx]), (None, kernel_size[0], kernel_size[1], 3, 2)],
                    amt_templates=self.output_dims[idx],
                    template_radius=self.template_radius,
                    activation="relu",
                    rotation_delta=self.rotation_deltas[idx]
                )
            )
            self.bn_layers.append(nn.BatchNorm1d(num_features=self.output_dims[idx]))
            self.amp_layers.append(AngularMaxPooling())

        #########
        # Output
        #########
        if segmentation_classes:
            self.output_dense = nn.Linear(in_features=self.output_dims[-1], out_features=segmentation_classes)
        else:
            self.output_dense = nn.Linear(in_features=self.output_dims[-1], out_features=6890)

    def forward(self, inputs):
        #################
        # Handling Input
        #################
        signal, bc = inputs
        signal = self.normalize(signal)
        signal = self.downsize_dense(signal)
        signal = self.downsize_activation(signal)
        signal = self.downsize_bn(signal)

        ###############
        # Forward pass
        ###############
        for idx in range(len(self.output_dims)):
            signal = self.do_layers[idx](signal)
            signal = self.isc_layers[idx]([signal, bc])
            signal = self.amp_layers[idx](signal)
            signal = self.bn_layers[idx](signal)

        #########
        # Output
        #########
        return self.output_dense(signal)

    def train_loop(self,
                   dataset,
                   loss_fn,
                   opt,
                   decay_rate=0.95,
                   decay_steps=500,
                   verbose=True,
                   epoch=None,
                   prev_steps=None,
                   use_lr_decay=False):
        self.train()
        epoch_accuracy = 0.
        epoch_loss = 0.
        mean_accuracy = 0.
        mean_loss = 0.

        for step, ((signal, bc), gt) in enumerate(dataset):
            pred = self([signal, bc])
            loss = loss_fn(pred, gt)
            opt.zero_grad()
            loss.backward()
            opt.step()

            if use_lr_decay:
                custom_exp_scheduler(opt, prev_steps + step, decay_rate=decay_rate, decay_steps=decay_steps)

            # Statistics
            epoch_accuracy = epoch_accuracy + multiclass_accuracy(pred, gt).detach()
            epoch_loss = epoch_loss + loss.detach()

            # I/O
            mean_accuracy = epoch_accuracy / (step + 1)
            mean_loss = epoch_loss / (step + 1)
            if verbose:
                sys.stdout.write(
                    f"\rEpoch: {epoch} - "
                    f"Training step: {step} - "
                    f"Loss {mean_loss:.4f} - "
                    f"Accuracy {mean_accuracy:.4f}"
                    f" - Memory: {print_mem()}"
                )

        return {"epoch_loss": mean_loss, "epoch_accuracy": mean_accuracy}

    def validation_loop(self, dataset, loss_fn, verbose=True):
        self.eval()

        with torch.no_grad():
            val_loss = 0.
            val_accuracy = 0.

            for step, ((signal, bc), gt) in enumerate(dataset):
                pred = self([signal, bc])

                # Statistics
                val_accuracy = val_accuracy + multiclass_accuracy(pred, gt).detach()
                val_loss = val_loss + loss_fn(pred, gt).detach()

            # I/O
            mean_accuracy = val_accuracy / (step + 1)
            mean_loss = val_loss / (step + 1)
            if verbose:
                sys.stdout.write(
                    f" - Val.-Loss: {mean_loss:.4f} - "
                    f"Val.-Accuracy: {mean_accuracy:.4f}"
                )
        return {"val_epoch_loss": mean_loss, "val_epoch_accuracy": mean_accuracy}



from geoconv.preprocessing.barycentric_coordinates import create_template_matrix

from abc import ABC, abstractmethod

import tensorflow as tf
import keras


class ConvIntrinsic(ABC, keras.layers.Layer):
    """A metaclass for intrinsic surface convolutions on Riemannian manifolds.

    Attributes
    ----------
    given_name: str

    activation_fn: str
        The activation function to use.
    rotation_delta: int
        The distance between two rotations. If `n` angular coordinates are given in the data, then the default behavior
        is to rotate the template `n` times, i.e. a shift in every angular coordinate of 1 to the next angular
        coordinate.
        If `rotation_delta = 2`, then the shift increases to 2 and the total amount of rotations reduces to
        ceil(n / rotation_delta). This gives a speed-up and saves memory. However, quality of results might worsen.
    amt_templates: int
        The amount of templates to apply during one convolution.
    template_radius: float
        The maximal geodesic extension of the template.
    template_regularizer: str or callable
        A regularizer for the template.
    bias_regularizer: str or callable
        A regularizer for the bias.
    initializer: str or callable
        An initializer for the template and bias.
    include_prior: bool
        Whether to weight the interpolations according to a pre-defined kernel.
    """

    def __init__(self,
                 amt_templates,
                 template_radius,
                 include_prior=True,
                 activation="relu",
                 rotation_delta=1,
                 name=None,
                 template_regularizer=None,
                 bias_regularizer=None,
                 initializer="glorot_uniform"):
        if name:
            super().__init__(name=name)
        else:
            super().__init__()

        self.given_name = name
        self.activation_fn = activation
        self.rotation_delta = rotation_delta
        self.amt_templates = amt_templates
        self.template_radius = template_radius
        self.template_regularizer = template_regularizer
        self.bias_regularizer = bias_regularizer
        self.initializer = initializer
        self.include_prior = include_prior

        # Attributes that depend on the data and are set automatically in build
        self._activation = keras.layers.Activation(self.activation_fn)
        self._bias = None
        self._all_rotations = None
        self._template_size = None  # (#radial, #angular)
        self._template_vertices = None
        self._template_neighbor_weights = None
        self._template_self_weights = None
        self._kernel = None
        self._feature_dim = None

    def get_config(self):
        config = super(ConvIntrinsic, self).get_config()
        config.update(
            {
                "amt_templates": self.amt_templates,
                "template_radius": self.template_radius,
                "include_prior": self.include_prior,
                "activation": self.activation_fn,
                "rotation_delta": self.rotation_delta,
                "name": self.given_name,
                "template_regularizer": self.template_regularizer,
                "bias_regularizer": self.bias_regularizer,
                "initializer": self.initializer
            }
        )
        return config

    def build(self, input_shape):
        """Builds the layer by setting template and bias attributes

        Parameters
        ----------
        input_shape: (tensorflow.TensorShape, tensorflow.TensorShape)
            The shape of the signal and the shape of the barycentric coordinates.
        """
        signal_shape, barycentric_shape = input_shape

        # Configure template
        self._template_size = (barycentric_shape[1], barycentric_shape[2])
        self._all_rotations = self._template_size[1]
        self._template_vertices = tf.constant(
            create_template_matrix(self._template_size[0], self._template_size[1], radius=self.template_radius)
        )
        self._feature_dim = signal_shape[-1]

        # Configure trainable weights
        self._template_neighbor_weights = self.add_weight(
            name="neighbor_weights",
            shape=(self.amt_templates, self._template_size[0], self._template_size[1], signal_shape[1]),
            initializer=self.initializer,
            trainable=True,
            regularizer=self.template_regularizer
        )
        self._template_self_weights = self.add_weight(
            name="center_weights",
            shape=(self.amt_templates, 1, signal_shape[1]),
            initializer=self.initializer,
            trainable=True,
            regularizer=self.template_regularizer
        )
        self._bias = self.add_weight(
            name="conv_intrinsic_bias",
            shape=(self.amt_templates,),
            initializer=self.initializer,
            trainable=True,
            regularizer=self.bias_regularizer
        )

        # Configure kernel
        self._configure_kernel()

    @tf.function
    def call(self, inputs, orientations=None):
        """Computes intrinsic surface convolution on all vertices of a given mesh.

        Parameters
        ----------
        inputs: (tensorflow.Tensor, tensorflow.Tensor)
            The first tensor represents the signal defined on the manifold. It has size
            (n_vertices, feature_dim). The second tensor represents the barycentric coordinates. It has
            size (n_vertices, n_radial, n_angular, 3, 2).
        orientations: tensorflow.Tensor
            Contains an integer that tells how to rotate the data.

        Returns
        -------
        tensorflow.Tensor
            The geodesic convolution of the template with the signal on the object mesh in every given GPC-system.
            It has size (vertices, n_rotations, templates)
        """
        mesh_signal, bary_coordinates = inputs

        ######################################################
        # Fold center - conv_center: (vertices, 1, templates)
        ######################################################
        # Weight matrix : (templates, 1, input_dim)
        # Mesh signal   : (vertices, input_dim)
        # Result        : (vertices, 1, n_templates)
        conv_center = tf.einsum("tef,kf->ket", self._template_self_weights, mesh_signal)

        #####################################################################
        # Fold neighbors - conv_neighbor: (vertices, n_rotations, templates)
        #####################################################################
        # Call patch operator
        interpolations = self._patch_operator(mesh_signal, bary_coordinates)
        # Determine orientations
        if orientations is None:
            # No specific orientations given. Hence, compute for all orientations.
            orientations = tf.range(start=0, limit=self._all_rotations, delta=self.rotation_delta)

        def fold_neighbor(o):
            # Weight              : (templates, radial, angular, input_dim)
            # Mesh interpolations : (vertices, radial, angular, input_dim)
            # Result              : (vertices, templates)
            return tf.einsum(
                "traf,kraf->kt",
                self._template_neighbor_weights,
                tf.roll(interpolations, shift=o, axis=2)
            )

        # conv_neighbor: (vertices, n_rotations, templates)
        conv_neighbor = tf.transpose(
            tf.map_fn(fold_neighbor, orientations, fn_output_signature=tf.float32), perm=[1, 0, 2]
        )
        return self._activation(conv_center + conv_neighbor + self._bias)

    @tf.function
    def _patch_operator(self, mesh_signal, barycentric_coordinates):
        """Interpolates and weights mesh signal

        Parameters
        ----------
        mesh_signal: tensorflow.Tensor
            The signal values at the template vertices
        barycentric_coordinates: tensorflow.Tensor
            The barycentric coordinates for the template vertices

        Returns
        -------
        tensorflow.Tensor:
            Weighted and interpolated mesh signals
        """
        interpolations = self._signal_retrieval(mesh_signal, barycentric_coordinates)

        if self.include_prior:
            # Weight matrix  : (radial, angular, radial, angular)
            # interpolations : (vertices, radial, angular, input_dim)
            # Result         : (vertices, radial, angular, input_dim)
            return tf.einsum("raxy,kxyf->kraf", self._kernel, interpolations)
        else:
            return interpolations

    @tf.function
    def _signal_retrieval(self, mesh_signal, barycentric_coordinates):
        """Interpolates signals at template vertices

        Parameters
        ----------
        mesh_signal: tensorflow.Tensor
            The signal values at the template vertices
        barycentric_coordinates: tensorflow.Tensor
            The barycentric coordinates for the template vertices

        Returns
        -------
        tensorflow.Tensor:
            Interpolation values for the template vertices
        """
        vertex_indices = tf.reshape(
            tf.cast(barycentric_coordinates[:, :, :, :, 0], tf.int32), (-1, 1)
        )
        mesh_signal = tf.reshape(
            tf.gather_nd(mesh_signal, vertex_indices),
            (-1, self._template_size[0], self._template_size[1], 3, self._feature_dim)
        )
        # (vertices, n_radial, n_angular, input_dim)
        return tf.math.reduce_sum(
            tf.expand_dims(barycentric_coordinates[:, :, :, :, 1], axis=-1) * mesh_signal, axis=-2
        )

    def _configure_kernel(self):
        """Defines all necessary interpolation coefficient matrices for the patch operator."""
        self._kernel = tf.cast(
            self.define_kernel_values(self._template_vertices.numpy()), tf.float32
        )

    @abstractmethod
    def define_kernel_values(self, template_matrix):
        """Defines the kernel values for each template vertex.

        Parameters
        ----------
        template_matrix: np.ndarray
            An array of size [n_radial, n_angular, 2], which contains the positions of the template vertices in
            cartesian coordinates.

        Returns
        -------
        np.ndarray:
            An array of size [n_radial, n_angular, n_radial, n_angular], which contains the interpolation weights for
            the patch operator '[D(x)f](rho_in, theta_in)'
        """
        pass



from geoconv_examples.mpi_faust.tensorflow.faust_data_set import load_preprocessed_faust
from geoconv.tensorflow.layers import AngularMaxPooling
from geoconv.tensorflow.layers.conv_dirac import ConvDirac
from geoconv.tensorflow.layers import ConvGeodesic
from geoconv.tensorflow.layers.conv_zero import ConvZero

import keras_tuner
import tensorflow as tf
import os
import keras


class HyperModel(keras_tuner.HyperModel):

    def __init__(self, signal_dim, kernel_size, template_radius, splits, layer_conf=None, variant="dirac"):
        super().__init__()
        self.signal_dim = signal_dim
        self.kernel_size = kernel_size
        self.template_radius = template_radius
        self.splits = splits

        if variant == "dirac":
            self.layer_type = ConvDirac
        elif variant == "geodesic":
            self.layer_type = ConvGeodesic
        elif variant == "zero":
            self.layer_type = ConvZero
        else:
            raise RuntimeError("Select a layer type from: ['dirac', 'geodesic', 'zero']")

        if layer_conf is None:
            self.output_dims = [96, 256, 384, 384, 256]
            self.rotation_deltas = [1 for _ in range(len(self.output_dims))]
        else:
            self.output_dims, self.rotation_deltas = list(zip(*layer_conf))

        self.normalize = keras.layers.Normalization(axis=-1, name="input_normalization")

    def build(self, hp):
        amp = AngularMaxPooling()
        signal_input = keras.layers.Input(shape=self.signal_dim, name="Signal_input")
        bc_input = keras.layers.Input(shape=(self.kernel_size[0], self.kernel_size[1], 3, 2), name="BC_input")

        #################
        # Handling Input
        #################
        signal = self.normalize(signal_input)
        # signal = keras.layers.Dense(64, activation="relu", name="Downsize")(signal)
        # signal = keras.layers.BatchNormalization(axis=-1, name="BN_downsize")(signal)

        #######################
        # Network Architecture
        #######################
        for idx in range(len(self.output_dims)):
            signal = keras.layers.Dropout(rate=0.2)(signal)
            signal = self.layer_type(
                amt_templates=self.output_dims[idx],
                template_radius=self.template_radius,
                activation="relu",
                name=f"ISC_layer_{idx}",
                rotation_delta=self.rotation_deltas[idx]
            )([signal, bc_input])
            signal = amp(signal)
            signal = keras.layers.BatchNormalization(axis=-1, name=f"BN_layer_{idx}")(signal)

        #########
        # Output
        #########
        output = keras.layers.Dense(6890, name="Output")(signal)

        ################
        # Compile Model
        ################
        init_lr = 0.0004607
        init_wd = 0.0052094
        model = keras.Model(inputs=[signal_input, bc_input], outputs=[output])
        loss = keras.losses.SparseCategoricalCrossentropy(from_logits=True)
        opt = keras.optimizers.AdamW(
            learning_rate=keras.optimizers.schedules.ExponentialDecay(
                initial_learning_rate=hp.Float(
                    "init_lr",
                    min_value=init_lr - .25 * init_lr,
                    max_value=init_lr + .25 * init_lr
                ),
                decay_steps=500,
                decay_rate=0.99
            ),
            weight_decay=hp.Float(
                "weight_decay",
                min_value=init_wd - .25 * init_wd,
                max_value=init_wd + .25 * init_wd
            )
        )
        model.compile(optimizer=opt, loss=loss, metrics=["sparse_categorical_accuracy"])
        return model


def hypertune(logging_dir,
              preprocessed_data,
              signal_dim,
              n_radial,
              n_angular,
              template_radius,
              splits,
              layer_conf=None):
    """Tunes the learning rate of the above IMCNN.

    Parameters
    ----------
    logging_dir: str
        The path to the logging directory. If nonexistent, directory will be created.
    preprocessed_data: str
        The path to the preprocessed data (the *.zip-file).
    signal_dim:
        The dimensionality of the signal.
    n_radial:
        The amount of radial coordinates of the kernel.
    n_angular:
        The amount of angular coordinates of the kernel.
    template_radius:
        The template radius.
    splits:
        The amount of splits for the ISC-layers.
    layer_conf: list
        List of tuples: The first entry references the output dimensions of the i-th ISC-layer, The second
        entry references of skips between each rotation while computing the convolution (rotation delta).
    model: str
        [OPTIONAL] If model == "resnet": ISCResnet18 will be trained, otherwise model from model.py.
    """
    # Create logging dir if necessary
    if not os.path.exists(logging_dir):
        os.makedirs(logging_dir)

    # Load data
    preprocess_zip = f"{preprocessed_data}.zip"
    kernel_size = (n_radial, n_angular)
    train_data = load_preprocessed_faust(preprocess_zip, signal_dim=signal_dim, kernel_size=kernel_size, set_type=0)
    val_data = load_preprocessed_faust(preprocess_zip, signal_dim=signal_dim, kernel_size=kernel_size, set_type=1)

    # Load hypermodel
    hyper = HyperModel(
        signal_dim=signal_dim,
        kernel_size=kernel_size,
        template_radius=template_radius,
        splits=splits,
        layer_conf=layer_conf
    )

    # Adapt normalization
    print("Initializing normalization layer..")
    hyper.normalize.build(tf.TensorShape([6890, signal_dim]))
    adaption_data = load_preprocessed_faust(
        preprocess_zip, signal_dim=signal_dim, kernel_size=kernel_size, set_type=0, only_signal=True
    )
    hyper.normalize.adapt(adaption_data)
    print("Done.")

    # Configure tuner
    tuner = keras_tuner.Hyperband(
        hypermodel=hyper,
        objective=[
            keras_tuner.Objective("val_loss", "min"),
            keras_tuner.Objective("val_sparse_categorical_accuracy", "max")
        ],
        max_epochs=200,
        directory=f"{logging_dir}/keras_tuner",
        project_name=f"faust_example",
        seed=42
    )
    # tuner = keras_tuner.BayesianOptimization(
    #     hypermodel=hyper,
    #     objective=[
    #         keras_tuner.Objective("val_loss", "min"),
    #         keras_tuner.Objective("val_sparse_categorical_accuracy", "max")
    #     ],
    #     max_trials=1000,
    #     directory=f"{logging_dir}/keras_tuner",
    #     project_name=f"faust_example",
    #     seed=42
    # )

    # Start hyperparameter-search
    stop = keras.callbacks.EarlyStopping(monitor="val_loss", patience=20)
    tuner.search(
        x=train_data.prefetch(tf.data.AUTOTUNE),
        validation_data=val_data.prefetch(tf.data.AUTOTUNE),
        callbacks=[stop]
    )
    print(tuner.results_summary())

    # Save best model
    best_model = tuner.get_best_models(num_models=1)[0]
    best_model.build(input_shape=[(signal_dim,), (n_radial, n_angular, 3, 2)])
    print(best_model.summary())
    best_model.save(f"{logging_dir}/best_model")



from geoconv_examples.mpi_faust.data.preprocess_faust import get_file_number

from torch.utils.data import IterableDataset

import torch
import random
import numpy as np
import os


def faust_generator(path_to_zip,
                    set_type=0,
                    only_signal=False,
                    device=None,
                    return_coordinates=False,
                    set_indices=None):
    """Reads one element of preprocessed FAUST-geoconv_examples into memory per 'next'-call.

    Parameters
    ----------
    path_to_zip: str
        The path to the .zip-file that contains the preprocessed faust data set
    set_type: int
        This integer has to be either:
            - 0 -> "train"  (adds noise to barycentric coordinates)
            - 1 -> "validation"
            - 2 -> "test"
            - 3 -> "all meshes"
        Depending on the choice, the training-, validation or testing data set will be returned. The split is equal to
        the one given in:
        > [Geodesic Convolutional Neural Networks on Riemannian Manifolds](https://arxiv.org/abs/1501.06297)
        > Jonathan Masci and Davide Boscaini et al.
    only_signal: bool
        Return only the signal matrices.
    device:
        The device to put the data on.
    return_coordinates: bool
        Whether to return the coordinates of the mesh vertices. Requires coordinates to be contained in preprocessed
        dataset.
    set_indices: list
        A list of integer values that determine which meshes shall be returned. If it is set to 'None', the set
        type determine which meshes will be returned. Defaults to 'None'. Adds noise to barycentric coordinates
        if set type is set to 0.

    Returns
    -------
    generator:
        A generator yielding the preprocessed data. I.e. the signal defined on the vertices, the barycentric coordinates
        and the ground truth correspondences.
    """
    # Initialize and sort file names
    dataset = np.load(path_to_zip, allow_pickle=True)
    file_names = [os.path.basename(fn) for fn in dataset.files]
    SIGNAL = [file_name for file_name in file_names if file_name.startswith("SIGNAL")]
    BC = [file_name for file_name in file_names if file_name.startswith("BC")]
    GT = [file_name for file_name in file_names if file_name.startswith("GT")]
    SIGNAL.sort(key=get_file_number), BC.sort(key=get_file_number), GT.sort(key=get_file_number)
    if return_coordinates:
        COORD = [file_name for file_name in file_names if file_name.startswith("COORD")]
        COORD.sort(key=get_file_number)

    # Set iteration indices according to set type
    if set_indices is None:
        if set_type == 0:
            indices = list(range(70))
            random.shuffle(indices)
        elif set_type == 1:
            indices = range(70, 80)
        elif set_type == 2:
            indices = range(80, 100)
        elif set_type == 3:
            indices = range(100)
        else:
            raise RuntimeError(
                f"There is no 'set_type'={set_type}. Choose from: [0: 'train', 1: 'val', 2: 'test', 3: 'all']."
            )
    else:
        indices = set_indices

    for idx in indices:
        # Read signal
        signal = torch.tensor(dataset[SIGNAL[idx]], dtype=torch.float32)

        # Read bc + add noise
        bc = torch.tensor(dataset[BC[idx]], dtype=torch.float32)
        kernel_size = bc.shape[1:3]

        if set_type == 0:
            noise = np.abs(np.random.normal(size=(bc.shape[0],) + kernel_size + (3, 2), scale=1e-5))
            noise[:, :, :, :, 0] = 0
            bc = bc + noise

        # Ground truth: Return the indices of the ones for each row
        gt = torch.tensor(dataset[GT[idx]], dtype=torch.int64).view(-1,)

        if device:
            if only_signal:
                yield signal.to(device)
            else:
                yield (signal.to(device), bc.to(device)), gt.to(device)
        else:
            if only_signal:
                yield signal
            else:
                # Coordinates are not required during training. However, other applications might need them.
                if return_coordinates:
                    coord = torch.tensor(dataset[COORD[idx]], dtype=torch.float32)
                    yield (signal, bc, coord), gt
                else:
                    yield (signal, bc), gt


class FaustDataset(IterableDataset):
    def __init__(self, path_to_zip, set_type=0, only_signal=False, device=None, return_coordinates=False):
        self.only_signal = only_signal
        self.path_to_zip = path_to_zip
        self.set_type = set_type
        self.only_signal = only_signal
        self.return_coordinates = return_coordinates
        self.device = device
        self.dataset = faust_generator(
            self.path_to_zip,
            set_type=self.set_type,
            only_signal=self.only_signal,
            device=self.device,
            return_coordinates=self.return_coordinates
        )

    def __iter__(self):
        return self.dataset

    def reset(self):
        self.dataset = faust_generator(
            self.path_to_zip,
            set_type=self.set_type,
            only_signal=self.only_signal,
            device=self.device,
            return_coordinates=self.return_coordinates
        )


from geoconv.pytorch.layers.conv_intrinsic import ConvIntrinsic

import numpy as np
import scipy as sp


def angle_distance(theta_max, theta_min):
    return np.minimum(theta_max - theta_min, theta_min + 2. * np.pi - theta_max)


def normal_pdf(mean_rho, mean_theta, var_rho, var_theta, rho, theta):
    """Normal probability distribution for geodesic polar coordinates

    Parameters
    ----------
    mean_rho: float
        Mean radial distance of the normal
    mean_theta: float
        Mean angle for the Gaussian
    var_rho: float
        Mean radial distance variance of the kernel vertices
    var_theta: float
        Mean angle variance of the kernel vertices
    rho: float
        Radial coordinate of the interpolation point that shall be weighted
    theta: float
        Angular coordinate of the interpolation point that shall be weighted

    Returns
    -------
    float:
        The weight for the interpolation point (rho, theta)
    """
    norm_coefficient = 1 / np.sqrt((2 * np.pi) ** 2 * var_rho * var_theta)
    max_angle = np.maximum(mean_theta, theta)
    min_angle = np.minimum(mean_theta, theta)
    delta_angle = angle_distance(max_angle, min_angle)
    vec = np.array([rho - mean_rho, delta_angle])
    mat = np.array([[1 / var_rho, 0], [0, 1 / var_theta]])
    exp = np.exp(-(1 / 2) * vec.T @ mat @ vec)
    return norm_coefficient * exp


class ConvGeodesic(ConvIntrinsic):
    """The geodesic convolutional layer

    Paper, that introduced the geodesic convolution:
    > [Geodesic Convolutional Neural Networks on Riemannian Manifolds](https://arxiv.org/abs/1501.06297)
    > Jonathan Masci and Davide Boscaini et al.
    """

    def define_kernel_values(self, template_matrix):
        interpolation_coefficients = np.zeros(template_matrix.shape[:-1] + template_matrix.shape[:-1])
        var_rho = template_matrix[:, :, 0].var()
        var_theta = template_matrix[:, :, 1].var()
        for mean_rho_idx in range(template_matrix.shape[0]):
            for mean_theta_idx in range(template_matrix.shape[1]):
                mean_rho, mean_theta = template_matrix[mean_rho_idx, mean_theta_idx]
                for rho_idx in range(template_matrix.shape[0]):
                    for theta_idx in range(template_matrix.shape[1]):
                        rho, theta = template_matrix[rho_idx, theta_idx]
                        interpolation_coefficients[mean_rho_idx, mean_theta_idx, rho_idx, theta_idx] = normal_pdf(
                            mean_rho, mean_theta, var_rho, var_theta, rho, theta
                        )
                interpolation_coefficients[mean_rho_idx, mean_theta_idx] = sp.special.softmax(
                    interpolation_coefficients[mean_rho_idx, mean_theta_idx]
                )
        return interpolation_coefficients



from geoconv.preprocessing.barycentric_coordinates import polar_to_cart
from geoconv.utils.misc import get_neighbors, get_faces_of_edge, compute_vector_angle, gpc_systems_into_cart

from matplotlib import pyplot as plt
from matplotlib.patches import Polygon

import c_extension
import numpy as np
import sys


class GPCSystem:
    def __init__(self, source_point, object_mesh, use_c=True, soft_clear=False):
        """Compute the initial radial and angular coordinates around a source point and setup caches.

        Angle coordinates are always given w.r.t. some reference direction. The choice of a reference
        direction can be arbitrary. Here, we choose the vector `x - source_point` with `x` being the
        first neighbor return by `get_neighbors` as the reference direction.

        This class handles two caches which connect GPC-system- with mesh information:
            - Edge-cache: Remembers all edges to a vertex
            - Face-cache: Remembers all faces to a sorted edge

        Parameters
        ----------
        source_point: int
            The index of the source point around which a window (GPC-system) shall be established.
        object_mesh: trimesh.Trimesh
            A loaded object mesh.
        use_c: bool
            A flag whether to use the c-extension.
        soft_clear: bool
            Whether to re-use old edge- and face-caches.
        """
        # Remember the underlying mesh
        self.object_mesh = object_mesh
        self.source_point = source_point
        self.radial_coordinates = np.full((object_mesh.vertices.shape[0],), np.inf)
        self.angular_coordinates = np.full((object_mesh.vertices.shape[0],), -1.0)
        self.x_coordinates = np.full((object_mesh.vertices.shape[0],), np.inf)
        self.y_coordinates = np.full((object_mesh.vertices.shape[0],), np.inf)

        #######################################
        # Calculate initial radial coordinates
        #######################################
        source_point_neighbors = get_neighbors(source_point, object_mesh)
        r3_source_point = object_mesh.vertices[source_point]
        r3_neighbors = object_mesh.vertices[source_point_neighbors]
        self.radial_coordinates[source_point_neighbors] = np.linalg.norm(
            r3_neighbors - np.stack([r3_source_point for _ in range(len(source_point_neighbors))]), ord=2, axis=-1
        )
        self.radial_coordinates[source_point] = .0

        ########################################
        # Calculate initial angular coordinates
        ########################################
        ref_neighbor = source_point_neighbors[0]
        rotation_axis = object_mesh.vertex_normals[source_point]
        theta_neighbors = np.full((len(source_point_neighbors, )), .0)
        for idx, neighbor in enumerate(source_point_neighbors):
            vector_a = object_mesh.vertices[ref_neighbor] - object_mesh.vertices[source_point]
            vector_b = object_mesh.vertices[neighbor] - object_mesh.vertices[source_point]
            if use_c:
                theta_neighbors[idx] = c_extension.compute_angle_360(vector_a, vector_b, rotation_axis)
            else:
                theta_neighbors[idx] = compute_vector_angle(vector_a, vector_b, rotation_axis)
        self.angular_coordinates[source_point_neighbors] = theta_neighbors
        self.angular_coordinates[source_point] = 0.0

        ##########################################
        # Calculate initial Cartesian coordinates
        ##########################################
        self.x_coordinates[source_point] = 0.
        self.y_coordinates[source_point] = 0.
        for neighbor in source_point_neighbors:
            x, y = polar_to_cart(angles=self.angular_coordinates[neighbor], scales=self.radial_coordinates[neighbor])
            self.x_coordinates[neighbor] = x
            self.y_coordinates[neighbor] = y

        ####################################################################################
        # Initialize face- and edge-cache with one-hop-neighborhood edges from source-point
        ####################################################################################
        if not soft_clear:
            self.edges = {-1: []}
            self.faces = {(-1, -1): []}
        self.edges[source_point] = []
        for neighbor in source_point_neighbors:
            edge, considered_faces = get_faces_of_edge(np.array([source_point, neighbor]), object_mesh)
            edge = list(edge)
            # Add edges to edge-cache
            self.add_edge(edge)
            # Add faces to face-cache
            for face in considered_faces:
                self.add_face(face)

    def soft_clear(self, source_point, use_c=True):
        """Reset radial- and angular coordinates, keep underlying mesh and edge- and face-caches.

        Parameters
        ----------
        source_point: int
            The new source point
        use_c: bool
            A flag whether to use the c-extension
        """
        self.__init__(source_point, self.object_mesh, use_c=use_c, soft_clear=True)

    def add_edge(self, edge):
        """Add an edge to the GPC-system

        Parameters
        ----------
        edge: list
            The edge to add
        """
        if np.inf in [self.x_coordinates[edge[0]], self.x_coordinates[edge[1]]]:
            raise RuntimeError(f"Edge {edge} lacks GPC: {[self.x_coordinates[edge[0]], self.x_coordinates[edge[1]]]}")

        edge = list(np.sort(edge))
        # Check if edge was seen once
        if edge not in self.edges[-1]:
            self.edges[-1].append(edge)
        for vertex in edge:
            # Check whether node already exists
            if vertex not in self.edges.keys():
                self.edges[vertex] = []
            # Check whether edge is already saved under node
            if edge not in self.edges[vertex]:
                self.edges[vertex].append(edge)

    def add_face(self, face):
        """Add a face to the GPC-system

        Parameters
        ----------
        face: np.ndarray
            The face to add
        """
        face = list(np.sort(face))
        if face not in self.faces[(-1, -1)]:
            self.faces[(-1, -1)].append(face)
        face_edges = [
            [face[0], face[1]], [face[1], face[2]], [face[0], face[2]]
        ]
        for edge in face_edges:
            if (edge[0], edge[1]) not in self.faces.keys():
                self.faces[(edge[0], edge[1])] = [face]
            elif face not in self.faces[(edge[0], edge[1])]:
                self.faces[(edge[0], edge[1])].append(face)

            # Recursively check all edges on whether their 2nd face is entirely describable with GPCs
            for new_face in get_faces_of_edge(edge, self.object_mesh)[1]:
                new_face = np.sort(new_face)
                # If all face coordinates are known and face has not been seen, then update GPC-system with `new_face`
                if (not np.array_equal(new_face, face)
                    and not np.any(np.isinf(self.radial_coordinates[new_face]))
                    and list(new_face) not in self.faces[(-1, -1)]):
                    self.update(
                        new_face[0],
                        self.radial_coordinates[new_face[0]],
                        self.angular_coordinates[new_face[0]],
                        new_face[1],
                        [new_face[2]],
                        update=False
                    )

    def update(self, vertex_i, rho_i, theta_i, vertex_j, k_vertices, plot_name="", update=True):
        """Update the GPC-system while preventing to edge intersections

        Parameters
        ----------
        vertex_i: int
            The updated vertex
        rho_i: float
            The new radial coordinate for `vertex_i`
        theta_i: float
            The new angular coordinate for `vertex_i`
        vertex_j: int
            The first vertex that we used to update the coordinates of `vertex_i`
        k_vertices: list
            The list of second vertices that could have potentially been used to update the coordinates of `vertex_i`
        plot_name: string
            If given, the update step is plotted and stored under the given name.
        update: bool
            Whether to update the stored coordinates of `vertex_i`.

        Returns
        -------
        bool:
            Whether the update succeeded, i.e. the update on `vertex_i` did not cause intersections
        """
        for vertex_k in [k for k in k_vertices if not np.isinf(self.radial_coordinates[k])]:
            # Sort vertex indices such that edge-cache does not store edges twice
            sorted_face = np.sort([vertex_i, vertex_j, vertex_k])

            ###############################################################################################
            # Collect all edges of `vertex_i` that will be added or are captured by the current GPC-system
            ###############################################################################################
            updated_face_edges = [
                [sorted_face[0], sorted_face[1]], [sorted_face[1], sorted_face[2]], [sorted_face[0], sorted_face[2]]
            ]

            if vertex_i in self.edges.keys():
                edges_of_interest = self.edges[vertex_i].copy()
            else:
                self.edges[vertex_i] = []
                edges_of_interest = []

            for edge in updated_face_edges:
                if edge not in edges_of_interest:
                    edges_of_interest.append(edge)

            ##########################################
            # Check collected edges for intersections
            ##########################################
            x, y = polar_to_cart(angles=theta_i, scales=rho_i)
            for edge in edges_of_interest:
                if edge[0] == vertex_i:
                    edge_fst_vertex = [x, y]
                else:
                    edge_fst_vertex = [self.x_coordinates[edge[0]], self.y_coordinates[edge[0]]]

                if edge[1] == vertex_i:
                    edge_snd_vertex = [x, y]
                else:
                    edge_snd_vertex = [self.x_coordinates[edge[1]], self.y_coordinates[edge[1]]]

                if self.line_segment_intersection(edge_fst_vertex, edge_snd_vertex):
                    # Return 'False' to indicate failed update due to intersection
                    return False

            ###########################
            # Update GPC of `vertex_i`
            ###########################
            if update:
                self.radial_coordinates[vertex_i] = rho_i
                self.angular_coordinates[vertex_i] = theta_i

                self.x_coordinates[vertex_i] = x
                self.y_coordinates[vertex_i] = y

            ################
            # Add new edges
            ################
            for edge in edges_of_interest:
                self.add_edge(edge)

            ###############
            # Add new face
            ###############
            self.add_face(sorted_face)

        if plot_name:
            self.plot([vertex_i, vertex_j], plot_name)

        # Return 'True' to indicate successful update
        return True

    def plot(self, new_line, save_name):
        """Plots the currently captured faces and highlights the updated line and vertex.

        new_line: list
            The vertex indices of the updated line.
        save_name: str
            The path under which the plot is saved.
        """
        fig, ax = plt.subplots(1, 1)
        x_new_0 = self.x_coordinates[new_line[0]]
        y_new_0 = self.y_coordinates[new_line[0]]
        x_new_1 = self.x_coordinates[new_line[1]]
        y_new_1 = self.y_coordinates[new_line[1]]
        ax.plot([x_new_0, x_new_1], [y_new_0, y_new_1], color="blue", linewidth=2., label="Updated Line")
        ax.scatter([x_new_0], [y_new_0], color="blue", s=50., label="Updated Vertex")

        all_coordinates = set()
        for face in self.faces[(-1, -1)]:
            x0 = self.x_coordinates[face[0]]
            y0 = self.y_coordinates[face[0]]

            x1 = self.x_coordinates[face[1]]
            y1 = self.y_coordinates[face[1]]

            x2 = self.x_coordinates[face[2]]
            y2 = self.y_coordinates[face[2]]

            coordinates = [(x0, y0), (x1, y1), (x2, y2)]
            for c in coordinates:
                all_coordinates.add(c)

            polygon = Polygon(np.array(coordinates), alpha=.4, edgecolor="red")
            ax.add_patch(polygon)

        all_coordinates = np.array([list(c) for c in all_coordinates])
        ax.set_xlim(all_coordinates[:, 0].min(), all_coordinates[:, 0].max())
        ax.set_ylim(all_coordinates[:, 1].min(), all_coordinates[:, 1].max())

        plt.savefig(f"{save_name}.svg")
        plt.legend(loc="upper right")
        plt.close()

    def line_segment_intersection(self, edge_fst_vertex, edge_snd_vertex):
        """Checks, whether a line segment intersects previously existing ones.

        Implements:
        https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection#Given_two_points_on_each_line_segment

        Parameters
        ----------
        edge_fst_vertex: np.ndarray
            The cartesian coordinates of the first vertex of the line to check
        edge_snd_vertex: np.ndarray
            The cartesian coordinates of the second vertex of the line to check

        Returns
        -------
        bool:
            Whether the new line segment intersect already existing ones.
        """
        x1, y1 = edge_fst_vertex[0], edge_fst_vertex[1]
        x2, y2 = edge_snd_vertex[0], edge_snd_vertex[1]
        all_edges = np.array(self.edges[-1])
        xs3, ys3 = self.x_coordinates[all_edges[:, 0]], self.y_coordinates[all_edges[:, 0]]
        xs4, ys4 = self.x_coordinates[all_edges[:, 1]], self.y_coordinates[all_edges[:, 1]]

        denominators = (x1 - x2) * (ys3 - ys4) - (y1 - y2) * (xs3 - xs4)
        if 0. in denominators:
            denominators += sys.float_info.min
        nominators_1 = (x1 - xs3) * (ys3 - ys4) - (y1 - ys3) * (xs3 - xs4)
        nominators_2 = (x1 - xs3) * (y1 - y2) - (y1 - ys3) * (x1 - x2)

        xs = nominators_1 / denominators
        ys = nominators_2 / denominators

        eps = 1e-5
        return np.any(
            np.logical_and(np.logical_and(0. + eps < xs, xs < 1. - eps), np.logical_and(0. + eps < ys, ys < 1. - eps))
        )

    def get_gpc_system(self):
        """Return the GPC-system as one numpy array.

        An array `self.radial_coordinates` of radial coordinates from the source point to other points in the object
        mesh. An array `self.angular_coordinates` of angular coordinates of neighbors from `source_point` in its window.
        """
        return np.stack([self.radial_coordinates, self.angular_coordinates], axis=1)

    def get_gpc_triangles(self, in_cart=False):
        """Return all triangles captured by the current GPC-system.

        Parameters
        ----------
        in_cart: bool
            Whether to translate geodesic polar coordinates into cartesian.
        """
        gpc_system_triangles = self.get_gpc_system()
        gpc_system_triangles = gpc_system_triangles[np.array(self.faces[(-1, -1)])]
        if in_cart:
            return gpc_systems_into_cart(gpc_system_triangles)
        else:
            return gpc_system_triangles



from geoconv.pytorch.layers.conv_intrinsic import ConvIntrinsic

import numpy as np


class ConvDirac(ConvIntrinsic):
    """No interpolation weighting"""

    def __init__(self, *args, **kwargs):
        kwargs["include_prior"] = False  # Interpolation coefficients not required for Dirac prior
        super().__init__(*args, **kwargs)

    def define_kernel_values(self, template_matrix):
        """
        Only take the value at ('rho_in', 'theta_in') into account for the patch operator at ('rho_in', 'theta_in')

        [DEPRECATED]:
        This function is only implemented for visualization purposes. During the intrinsic surface convolution,
        interpolation coefficients of the Dirac prior do not need to be used as they do not alter the signal at the
        template vertices.
        """
        interpolation_coefficients = np.zeros(template_matrix.shape[:-1] + template_matrix.shape[:-1])
        for mean_rho_idx in range(template_matrix.shape[0]):
            for mean_theta_idx in range(template_matrix.shape[1]):
                interpolation_coefficients[mean_rho_idx, mean_theta_idx, mean_rho_idx, mean_theta_idx] = 1.
        return interpolation_coefficients



from geoconv.preprocessing.gpc_system import GPCSystem
from geoconv.preprocessing.gpc_system_utils import compute_distance_and_angle
from geoconv.utils.misc import get_neighbors

from multiprocessing import Pool
from tqdm import tqdm

import numpy as np
import warnings
import heapq


class GPCSystemGroup:
    def __init__(self, object_mesh, eps=0.000001, use_c=True, processes=1):
        self.object_mesh = object_mesh
        self.eps = eps
        self.use_c = use_c
        self.processes = processes
        self.object_mesh_gpc_systems = None

    def compute(self, u_max=.04):
        """Computes geodesic polar coordinates for all vertices within an object mesh.

        Parameters
        ----------
        u_max: float
            The maximal radius for each GPC-system.
        """
        n_vertices = self.object_mesh.vertices.shape[0]
        vertex_indices = np.arange(n_vertices)
        with Pool(self.processes) as p:
            gpc_systems = p.starmap(
                self.compute_gpc_system,
                tqdm(
                    [(vi, u_max) for vi in vertex_indices],
                    total=n_vertices,
                    postfix="Computing GPC-systems"
                )
            )
        self.object_mesh_gpc_systems = np.array(gpc_systems).flatten()

    def compute_gpc_system(self, source_point, u_max, gpc_system=None, plot_path=""):
        """Computes local GPC for one given source point.

        This method implements the algorithm of:
        > [Geodesic polar coordinates on polygonal meshes]
          (https://onlinelibrary.wiley.com/doi/full/10.1111/j.1467-8659.2012.03187.x)
        > Melvær, Eivind Lyche, and Martin Reimers.

        Parameters
        ----------
        source_point: int
            The index of the source point around which a window (GPC-system) shall be established
        u_max: float
            The maximal distance (e.g. radius of the patch) which a vertex may have to `source_point`
        gpc_system: GPCSystem
            A GPC-system that has been computed previously on the same mesh. Its caches can be re-used
            which saves computation time.
        plot_path: bool
            If given, the update steps will be plotted and stored at the given path.

        Returns
        -------
        GPCSystem:
            A GPC-system.
        """
        ########################
        # Initialize GPC-system
        ########################
        if gpc_system is None:
            gpc_system = GPCSystem(source_point, self.object_mesh, use_c=True)
        else:
            gpc_system.soft_clear(source_point)
        # Check whether initialization distances are larger than given max-radius
        check_array = np.array([x for x in gpc_system.radial_coordinates if not np.isinf(x)])
        if check_array.max() > u_max:
            warnings.warn(
                f"You chose a 'u_max' to be smaller then {check_array.max()}, which has been seen as an initialization"
                f" length for a GPC-system. Current GPC-system will only contain initialization vertices.",
                RuntimeWarning
            )

        ############################################
        # Initialize min-heap over radial distances
        ############################################
        candidates = []
        for neighbor in get_neighbors(source_point, self.object_mesh):
            candidates.append((gpc_system.radial_coordinates[neighbor], neighbor))
        heapq.heapify(candidates)

        ###################################
        # Algorithm to compute GPC-systems
        ###################################
        plot_number = 0
        while candidates:
            # Get vertex from min-heap that is closest to GPC-system origin
            j_dist, j = heapq.heappop(candidates)
            j_neighbors = get_neighbors(j, self.object_mesh)
            j_neighbors = [j for j in j_neighbors if j != source_point]
            for i in j_neighbors:
                # Compute the (updated) geodesic distance `new_u_i` and angular coordinate of the i-th neighbor from the
                # closest vertex in the min-heap to the source point of the GPC-system
                new_u_i, new_theta_i, k_vertices = compute_distance_and_angle(
                    i,
                    j,
                    gpc_system,
                    self.use_c,
                    rotation_axis=self.object_mesh.vertex_normals[source_point]
                )
                # In difference to the original pseudocode, we add 'new_u_i < u_max' to this IF-query
                # to ensure that the radial coordinates do not exceed 'u_max'.
                if new_u_i < u_max and gpc_system.radial_coordinates[i] / new_u_i > 1 + self.eps:
                    if plot_path:
                        if gpc_system.update(
                            i, new_u_i, new_theta_i, j, k_vertices, plot_name=f"{plot_path}/{plot_number}"
                        ):
                            heapq.heappush(candidates, (new_u_i, i))
                            plot_number += 1
                    else:
                        if gpc_system.update(i, new_u_i, new_theta_i, j, k_vertices):
                            heapq.heappush(candidates, (new_u_i, i))
        return gpc_system



from tqdm import tqdm

import numpy as np
import sys


def compute_barycentric(query_vertex, triangle):
    """Computes barycentric coordinates

    Compare: https://blackpawn.com/texts/pointinpoly/

    Parameters
    ----------
    query_vertex: np.ndarray
        1D-array that contains query-vertex in cartesian coordinates
    triangle: np.ndarray
        2D-array that depicts a triangle in cartesian coordinates

    Returns
    -------
    ((np.float32, np.float32, np.float32), bool)
        A tuple containing a triple and a boolean. The triple contains the barycentric coordinates for the vertices of
        the triangle. The boolean tells whether the query-vertex is within the triangle.
    """

    v0 = triangle[2] - triangle[0]
    v1 = triangle[1] - triangle[0]
    v2 = query_vertex - triangle[0]

    dot00, dot01, dot02 = v0.dot(v0), v0.dot(v1), v0.dot(v2)
    dot11, dot12 = v1.dot(v1), v1.dot(v2)

    denominator = dot00 * dot11 - dot01 * dot01
    if denominator == 0:
        denominator += sys.float_info.min
    point_2_weight = (dot11 * dot02 - dot01 * dot12) / denominator
    point_1_weight = (dot00 * dot12 - dot01 * dot02) / denominator
    point_0_weight = 1 - point_2_weight - point_1_weight

    is_inside_triangle = point_2_weight > 0 and point_1_weight > 0 and point_2_weight + point_1_weight <= 1

    return (point_0_weight, point_1_weight, point_2_weight), is_inside_triangle


def interpolation(query_point, gpc_triangles, gpc_triangles_node_indices):
    """Interpolates a query point within a GPC-system

    Parameters
    ----------
    query_point: np.ndarray
        A query point in cartesian coordinates
    gpc_triangles: np.ndarray
        The triangles contained in the GPC-system
    gpc_triangles_node_indices: np.ndarray
        The indices of the triangles contained in the GPC-system

    Returns
    -------
    (np.ndarray, np.ndarray):
        The first returned array contains the barycentric coordinates. The second returned array contains the
        indices of the vertices to which the barycentric coordinates belong. In case of 'vertex' not falling into
        any triangle, two zero arrays are returned. I.e. there will be no feature taken into consideration for this
        template vertex.
    """
    for idx, triangle in enumerate(gpc_triangles):
        b_coordinates, lies_within = compute_barycentric(query_point, triangle)
        if lies_within:
            return b_coordinates, gpc_triangles_node_indices[idx]
    return np.array([0., 0., 0.]), np.array([0, 0, 0])


def polar_to_cart(angles, scales=1.):
    """Returns x and y for a given angle.

    Parameters
    ----------
    angles: np.ndarray
        The angular coordinates
    scales: np.ndarray
        The radial coordinates

    Returns
    -------
    np.ndarray:
        The x-coordinate and y-coordinates

    """
    return np.stack([scales * np.cos(angles), scales * np.sin(angles)], axis=-1)


def create_template_matrix(n_radial, n_angular, radius, in_cart=False):
    """Creates a template matrix with radius `radius` and `n_radial` radial- and `n_angular` angular coordinates.

    Parameters
    ----------
    n_radial: int
        The amount of radial coordinates
    n_angular: int
        The amount of angular coordinates
    radius: float
        The radius of the template
    in_cart: bool
        If True, then the template matrix contains cartesian coordinates

    Returns
    -------
    np.ndarray
        A template matrix K with K[i, j] containing polar coordinates (radial, angular) of point (i. j)
    """

    coordinates = np.zeros((n_radial, n_angular, 2))
    for j in range(1, n_radial + 1):
        radial_coordinate = (j * radius) / n_radial
        for k in range(1, n_angular + 1):
            angular_coordinate = (2 * k * np.pi) / n_angular
            coordinates[j - 1, k - 1, 0] = radial_coordinate
            coordinates[j - 1, k - 1, 1] = angular_coordinate

    if in_cart:
        for rc in range(coordinates.shape[0]):
            for ac in range(coordinates.shape[1]):
                coordinates[rc, ac] = polar_to_cart(coordinates[rc, ac, 1], coordinates[rc, ac, 0])

    return coordinates


def compute_barycentric_coordinates(gpc_systems, n_radial=2, n_angular=4, radius=0.05):
    """Compute the barycentric coordinates for the given GPC-systems

    Parameters
    ----------
    gpc_systems: GPCSystemGroup
        The GPC-system-group for the underlying mesh
    n_radial: int
        The amount of radial coordinates of the template you wish to use
    n_angular: int
        The amount of angular coordinates of the template you wish to use
    radius: float
        The radius of the template of the template you wish to use
    verbose: bool
        Whether to print progress on terminal

    Returns
    -------
    A 5D-array containing the Barycentric coordinates for each template vertex and each GPC-system. It has the following
    structure:
        B[a, b, c, d, e]:
            - a: References GPC-system centered in vertex `a` of object mesh `object_mesh`
            - b: References the b-th radial coordinate of the template
            - c: References the c-th angular coordinate of the template
            - B[a, b, c, :, 0]: Returns the **indices** of the nodes that construct the triangle containing the template
                                vertex (b, c) in GPC-system centered in node `a`
            - B[a, b, c, :, 1]: Returns the **barycentric coordinates** of the nodes that construct the triangle
                                containing the template vertex (b, c) in GPC-system centered in node `a`
    """

    # Define template vertices at which interpolation values will be needed
    template_matrix = create_template_matrix(n_radial=n_radial, n_angular=n_angular, radius=radius, in_cart=True)
    n_gpc_systems = gpc_systems.object_mesh_gpc_systems.shape[0]
    barycentric_coordinates = np.zeros((n_gpc_systems, n_radial, n_angular, 3, 2))

    for gpc_system_idx in tqdm(range(n_gpc_systems), postfix=f"Computing barycentric coordinates"):
        gpc_system = gpc_systems.object_mesh_gpc_systems[gpc_system_idx]
        gpc_triangles = gpc_system.get_gpc_triangles(in_cart=True)
        for radial_coordinate in range(n_radial):
            for angular_coordinate in range(n_angular):
                bc, indices = interpolation(
                    template_matrix[radial_coordinate, angular_coordinate],
                    gpc_triangles,
                    gpc_system.faces[(-1, -1)]
                )
                barycentric_coordinates[gpc_system_idx, radial_coordinate, angular_coordinate, :, 0] = indices
                barycentric_coordinates[gpc_system_idx, radial_coordinate, angular_coordinate, :, 1] = bc

    return barycentric_coordinates



from geoconv.preprocessing.barycentric_coordinates import polar_to_cart

from tqdm import tqdm
from scipy.linalg import blas

import pygeodesic.geodesic as geodesic
import numpy as np
import trimesh


def compute_vector_angle(vector_a, vector_b, rotation_axis):
    """Compute the angle between two vectors

    Parameters
    ----------
    vector_a: np.ndarray
        The first vector
    vector_b: np.ndarray
        The second vector
    rotation_axis: [np.ndarray, None]
        For angles in [0, 2*pi[ in the 3-dimensional space an "up"-direction is required. If `None` is passed an angle
        between [0, pi[ is returned.

    Returns
    -------
    float:
        The angle between `vector_a` and `vector_b`
    """
    vector_a = vector_a / blas.dnrm2(vector_a)
    vector_b = vector_b / blas.dnrm2(vector_b)
    angle = blas.ddot(vector_a, vector_b)
    if angle > 1.0:
        angle = 1.0
    elif angle < -1.0:
        angle = -1.0
    angle = np.arccos(angle)
    if rotation_axis is None:
        return angle
    else:
        cross_product = np.cross(vector_a, vector_b)
        opposite_direction = rotation_axis.dot(cross_product) < 0.0
        angle = 2 * np.pi - angle if opposite_direction else angle
        return angle


def get_faces_of_edge(edge, object_mesh):
    """Determine both faces of a given edge

    Parameters
    ----------
    edge: np.ndarray
        The edge for which the faces shall be returned.
    object_mesh: trimesh.Trimesh
        The underlying mesh.
    """
    edge = np.sort(edge)
    # 1.) Get the edge index of `sorted_edge` "in both ways", i.e. two indices for `sorted_edge`
    edge_indices = object_mesh.edges_sorted == edge
    edge_indices = np.where(np.logical_and(edge_indices[:, 0], edge_indices[:, 1]))
    # 2.) Get faces of `sorted_edge` by retrieving `face_indices` for the found `edge_indices`
    face_indices = object_mesh.edges_face[edge_indices]
    considered_faces = object_mesh.faces[face_indices]
    # 3.) Return sorted edge and corresponding faces
    return edge, considered_faces


def get_neighbors(vertex, object_mesh):
    """Calculates the one-hop neighbors of a vertex

    Parameters
    ----------
    vertex: int
        The index of the vertex for which the neighbor indices shall be computed
    object_mesh: trimesh.Trimesh
        An object mesh

    Returns
    -------
    list:
        A list of neighboring vertex-indices.
    """

    return list(object_mesh.vertex_adjacency_graph[vertex].keys())


def normalize_mesh(mesh, geodesic_diameter=None):
    """Center mesh and scale x, y and z dimension with '1/geodesic diameter'.

    Parameters
    ----------
    mesh: trimesh.Trimesh
        The triangle mesh, that shall be normalized
    geodesic_diameter: float
        The geodesic diameter. If not provided, this function will compute the geodesic diameter.

    Returns
    -------
    (trimesh.Trimesh, float):
        The normalized mesh and the geodesic diameter, with which the mesh was normalized
    """
    # Center mesh
    for dim in range(3):
        mesh.vertices[:, dim] = mesh.vertices[:, dim] - mesh.vertices[:, dim].mean()

    # Determine geodesic diameter
    if geodesic_diameter is None:
        distance_matrix, geodesic_diameter = compute_geodesic_diameter(mesh)

    # Scale mesh
    for dim in range(3):
        mesh.vertices[:, dim] = mesh.vertices[:, dim] * (1 / geodesic_diameter)
    print(f"-> Normalized with geodesic diameter: {geodesic_diameter}")

    return mesh, geodesic_diameter


def compute_geodesic_diameter(mesh):
    """Computes the geodesic diameter of a mesh.

    In case the mesh contains a pair of vertices which are not connected by a path, this
    function returns the largest geodesic distance that has been seen as the geodesic diameter.

    Parameters
    ----------
    mesh: trimesh.Trimesh
        The triangle mesh, for which the geodesic diameter shall be calculated.

    Returns
    -------
    (np.array, float):
        The distance matrix between all vertices of the mesh and the geodesic diameter of the mesh.
    """
    n_vertices = mesh.vertices.shape[0]
    distance_matrix = np.zeros((n_vertices, n_vertices))
    geoalg = geodesic.PyGeodesicAlgorithmExact(mesh.vertices, mesh.faces)
    for sp in tqdm(range(n_vertices), postfix=f"Calculating geodesic diameter.."):
        distances, _ = geoalg.geodesicDistances([sp], None)
        distance_matrix[sp] = distances
    return distance_matrix, distance_matrix[distance_matrix != np.inf].max()


def gpc_systems_into_cart(gpc_systems):
    """Translates the geodesic polar coordinates of given GPC-systems into cartesian

    Parameters
    ----------
    gpc_systems: np.ndarray
        A 3D-array containing all GPC-systems which shall be translated

    Returns
    -------
    np.ndarray:
        The same GPC-systems but in cartesian coordinates
    """
    gpc_systems_cart = np.copy(gpc_systems)
    return polar_to_cart(gpc_systems_cart[:, :, 1], gpc_systems_cart[:, :, 0])


def reconstruct_template(gpc_system, b_coordinates):
    """Reconstructs the template vertices with barycentric coordinates

    Parameters
    ----------
    gpc_system: np.ndarray
        A 2D-array that describes a GPC-system. I.e. 'gpc_system[i]' contains the
        geodesic polar coordinates (radial, angle) for the i-th vertex of the underlying
        object mesh.
    b_coordinates: np.ndarray
        Contains the barycentric coordinates from which the template shall be reconstructed.
    Returns
    -------
    np.ndarray:
        Cartesian template coordinates in the same format as returned by 'create_template_matrix'.

    """

    reconstructed_template = np.zeros((b_coordinates.shape[0], b_coordinates.shape[1], 2))
    for rc in range(b_coordinates.shape[0]):
        for ac in range(b_coordinates.shape[1]):
            # Get vertices
            vertex_indices = b_coordinates[rc, ac, :, 0].astype(np.int16)
            vertices = [(gpc_system[vertex_indices[idx], 0], gpc_system[vertex_indices[idx], 1]) for idx in range(3)]
            vertices = np.array([polar_to_cart(angles=y, scales=x) for x, y in vertices])

            # Interpolate vertices
            weights = b_coordinates[rc, ac, :, 1]
            reconstructed_template[rc, ac] = vertices.T @ weights
    return reconstructed_template


def shuffle_mesh_vertices(mesh, given_shuffle=None):
    """Shuffles the vertices of the mesh

    Parameters
    ----------
    mesh: trimesh.Trimesh
        The mesh from which you want to shuffle the vertices
    given_shuffle: np.ndarray
        A given shuffle of the vertices

    Returns
    -------
    (trimesh.Trimesh, np.ndarray, np.ndarray)
        The same mesh but with a different vertices order. Additionally, two arrays are returned. Both contain vertex
        indices. Given a vertex index 'idx', it holds that:

        mesh.vertices[idx] == shuffled_mesh.vertices[shuffle_map[idx]] == mesh.vertices[ground_truth[shuffle_map[idx]]]
    """
    ground_truth = np.arange(mesh.vertices.shape[0])
    if given_shuffle is None:
        np.random.shuffle(ground_truth)
    else:
        ground_truth = np.copy(given_shuffle)
    mesh_vertices = np.copy(mesh.vertices)[ground_truth]

    shuffle_map = []
    for vertex_idx in range(mesh.vertices.shape[0]):
        shuffle_map.append(np.where(ground_truth == vertex_idx)[0])
    shuffle_map = np.array(shuffle_map).flatten()

    mesh_faces = np.copy(mesh.faces)
    for face_idx in range(mesh.faces.shape[0]):
        mesh_faces[face_idx, 0] = shuffle_map[mesh.faces[face_idx, 0]]
        mesh_faces[face_idx, 1] = shuffle_map[mesh.faces[face_idx, 1]]
        mesh_faces[face_idx, 2] = shuffle_map[mesh.faces[face_idx, 2]]

    shuffled_mesh = trimesh.Trimesh(vertices=mesh_vertices, faces=mesh_faces)

    return shuffled_mesh, shuffle_map, ground_truth


def get_included_faces(object_mesh, gpc_system):
    """Retrieves face indices from GPC-system

    Parameters
    ----------
    object_mesh: trimesh.Trimesh
        The object mesh
    gpc_system: np.ndarray
        The considered GPC-system

    Returns
    -------
    list:
        The list of face IDs which are included in the GPC-system
    """
    included_face_ids = []

    # Determine vertex IDs that are included in the GPC-system
    gpc_vertex_ids = np.arange(gpc_system.shape[0])[gpc_system[:, 0] != np.inf]

    # Determine what faces are entirely contained within the GPC-system
    for face_id, face in enumerate(object_mesh.faces):
        counter = 0
        for vertex_id in face:
            counter = counter + 1 if vertex_id in gpc_vertex_ids else counter
        if counter == 3:
            included_face_ids.append(face_id)

    return included_face_ids


def get_points_from_polygons(polygons):
    """Returns the unique set of points given in a set of polygons

    Parameters
    ----------
    polygons: np.ndarray
        Set of polygons from which the set of unique points will be returned

    Returns
    -------
    np.ndarray
        The set of unique points
    """
    return np.unique(polygons.reshape((-1, 2)), axis=0)


def find_largest_one_hop_dist(object_mesh):
    """Finds the largest Euclidean distance from center vertex to a one-hop neighbor in a triangle mesh

    Returns
    -------
    float:
        The largest initialization distance from a center-vertex to a one-hop neighbor in the triangle mesh
    """
    all_edges = object_mesh.vertices[object_mesh.edges]
    return np.linalg.norm(all_edges[:, 0, :] - all_edges[:, 1, :], axis=-1).max()



from geoconv.utils.misc import get_included_faces, normalize_mesh

from matplotlib import pyplot as plt
from multiprocessing import Pool
from tqdm import tqdm

import pygeodesic.geodesic as geodesic
import trimesh
import numpy as np


def princeton_benchmark(imcnn,
                        test_dataset,
                        ref_mesh_path,
                        file_name,
                        normalize=True,
                        plot_title="Princeton Benchmark",
                        curve_label=None,
                        plot=True,
                        processes=1,
                        geodesic_diameter=None,
                        pytorch_model=False):
    """Plots the accuracy w.r.t. a gradually changing geodesic error

    Princeton benchmark has been introduced in:
    > [Blended intrinsic maps](https://doi.org/10.1145/2010324.1964974)
    > Vladimir G. Kim, Yaron Lipman and Thomas Funkhouser

    Parameters
    ----------
    imcnn:
        The Intrinsic Mesh CNN
    test_dataset: tensorflow.data.Dataset
        The test dataset on which to evaluate the Intrinsic Mesh CNN
    ref_mesh_path: str
        A path to the reference mesh
    file_name: str
        The file name under which to store the plot and the data (without file format ending!)
    normalize: bool
        Whether to normalize the reference mesh
    plot_title: str
        The title of the plot
    curve_label: str
        The name displayed in the plot legend
    plot: bool
        Whether to plot immediately.
    processes: int
        The amount of concurrent processes.
    geodesic_diameter: float
        The geodesic diameter of the reference mesh
    pytorch_model: bool
        Whether a pytorch model is given.
    """

    reference_mesh = trimesh.load_mesh(ref_mesh_path)
    if normalize:
        reference_mesh, _ = normalize_mesh(reference_mesh, geodesic_diameter=geodesic_diameter)

    mesh_number = 0
    for ((signal, barycentric), ground_truth) in test_dataset:
        if pytorch_model:
            prediction = np.array(imcnn([signal, barycentric]).cpu()).argmax(axis=1)
            ground_truth = ground_truth.cpu()
        else:
            prediction = np.array(imcnn([signal, barycentric])).argmax(axis=1)
        batched = [(data, reference_mesh) for data in np.stack([ground_truth, prediction], axis=-1)]
        with Pool(processes) as p:
            geodesic_errors = p.starmap(
                geodesic_alg_wrapper,
                tqdm(batched, total=len(batched), postfix=f"Computing Princeton benchmark for test mesh {mesh_number}")
            )
        mesh_number += 1

    ##########################
    # Sorting geodesic errors
    ##########################
    geodesic_errors = np.array(geodesic_errors)
    geodesic_errors.sort()
    amt_values = geodesic_errors.shape[0]
    arr = np.array([((i + 1) / amt_values, x) for (i, x) in zip(range(amt_values), geodesic_errors)])
    np.save(f"{file_name}.npy", arr)

    ###############################################################
    # One y-value per x-value: Take highest percentage per x-value
    ###############################################################
    unique_x_values = np.unique(arr[:, 1])
    unique_values = []
    for unique_x in unique_x_values:
        unique_values.append(arr[np.where(arr[:, 1] == unique_x)[0][-1]])
    unique_values = np.array(unique_values)

    ###########
    # Plotting
    ###########
    plt.plot(unique_values[:, 1], unique_values[:, 0], label=curve_label)
    plt.title(plot_title)
    plt.xlabel("geodesic error")
    plt.ylabel("% correct correspondences")
    if plot:
        plt.grid()
        plt.legend()
        plt.savefig(f"{file_name}.svg")
        plt.show()


def geodesic_alg_wrapper(ground_truth_and_prediction, reference_mesh):
    """A wrapper function for PyGeodesicAlgorithmExact

    Required, since 'geodesic.PyGeodesicAlgorithmExact' can't be directly used as an argument for 'Pool.starmap'.

    Parameters
    ----------
    ground_truth_and_prediction: np.ndarray
        Simple array with two entries. First entry is the index of the ground truth vertex.
        The second entry is the index of the predicted vertex.
    reference_mesh: trimesh.Trimesh
        The triangle mesh on which the geodesic distances will be calculated.

    Returns
    -------
    float:
        The geodesic distance between the ground truth and predicted vertex.
    """
    geoalg = geodesic.PyGeodesicAlgorithmExact(reference_mesh.vertices, reference_mesh.faces)
    gt, pred = ground_truth_and_prediction
    return geoalg.geodesicDistance(pred, gt)[0]


def kernel_coverage(object_mesh, gpc_system, bary_coordinates):
    """Quality measure for how much a kernel covers within a GPC-system

    Parameters
    ----------
    object_mesh: trimesh.Trimesh
        An object mesh
    gpc_system:
        The considered GPC-system
    bary_coordinates:
        The barycentric coordinates for the kernel for the considered GPC-system


    Returns
    -------
    float
        The ratio between faces in which a kernel vertex lies and the total
        amount of faces within the GPC-system (kernel coverage)
    """
    patch = get_included_faces(object_mesh, gpc_system)
    patch_faces = object_mesh.faces[patch]
    covered_faces = []
    for angular in range(bary_coordinates.shape[0]):
        for radial in range(bary_coordinates.shape[1]):
            face_of_kernel_vertex = bary_coordinates[angular, radial, :, 0]

            # Check in what face the kernel vertex lies
            arr = np.array(face_of_kernel_vertex == patch_faces)
            arr = np.logical_and(np.logical_and(arr[:, 0], arr[:, 1]), arr[:, 2])
            arr = np.where(arr)[0]
            if len(arr):
                # Add face index to known faces if previously unknown
                face_idx = patch[arr[0]]
                if face_idx not in covered_faces:
                    covered_faces.append(face_idx)

    # Return the percentage of how many triangles in the patch are regarded by the kernel
    return len(covered_faces) / len(patch)


def evaluate_kernel_coverage(object_mesh, gpc_systems, bary_coordinates, verbose=True):
    """Computes the average kernel coverage of a GPC-system

    Parameters
    ----------
    object_mesh: trimesh.Trimesh
        An object mesh
    gpc_systems: np.ndarray
        A 3D-array containing multiple GPC-systems
    bary_coordinates:
        The barycentric coordinates of the kernels layed on the given GPC-systems

    Returns
    -------
    float
        The average coverage rate of the kernels on the GPC-systems
    """
    assert gpc_systems.shape[0] == bary_coordinates.shape[0], \
        "You must provide similar amount of GPC-system and Barycentric coordinate arrays."

    coverages = []
    for idx in range(gpc_systems.shape[0]):
        coverage = kernel_coverage(object_mesh, gpc_systems[idx], bary_coordinates[idx])
        coverages.append(coverage)
    if verbose:
        print(coverages)
    return np.mean(coverages)



from geoconv.tensorflow.layers.angular_max_pooling import AngularMaxPooling
from geoconv.tensorflow.layers.conv_geodesic import ConvGeodesic
from geoconv.tensorflow.layers.conv_zero import ConvZero
from geoconv.tensorflow.layers.conv_dirac import ConvDirac

import tensorflow as tf
import keras


class Imcnn(keras.Model):
    def __init__(self,
                 signal_dim,
                 kernel_size,
                 template_radius,
                 layer_conf=None,
                 variant="dirac",
                 segmentation_classes=-1,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.signal_dim = signal_dim
        self.kernel_size = kernel_size
        self.template_radius = template_radius

        if variant == "dirac":
            self.layer_type = ConvDirac
        elif variant == "geodesic":
            self.layer_type = ConvGeodesic
        elif variant == "zero":
            self.layer_type = ConvZero
        else:
            raise RuntimeError("Select a layer type from: ['dirac', 'geodesic', 'zero']")

        if layer_conf is None:
            self.output_dims = [96, 256, 384, 384]
            self.rotation_deltas = [1 for _ in range(len(self.output_dims))]
        else:
            self.output_dims, self.rotation_deltas = list(zip(*layer_conf))

        #################
        # Handling Input
        #################
        self.normalize = keras.layers.Normalization(axis=-1, name="input_normalization")
        self.downsize_dense = keras.layers.Dense(64, activation="relu", name="downsize")
        self.downsize_bn = keras.layers.BatchNormalization(axis=-1, name="BN_downsize")

        #############
        # ISC Layers
        #############
        self.isc_layers = []
        self.bn_layers = []
        self.do_layers = []
        self.amp_layers = []
        for idx in range(len(self.output_dims)):
            self.do_layers.append(keras.layers.Dropout(rate=0.2, name=f"DO_layer_{idx}"))
            self.isc_layers.append(
                self.layer_type(
                    amt_templates=self.output_dims[idx],
                    template_radius=self.template_radius,
                    activation="relu",
                    name=f"ISC_layer_{idx}",
                    rotation_delta=self.rotation_deltas[idx]
                )
            )
            self.bn_layers.append(keras.layers.BatchNormalization(axis=-1, name=f"BN_layer_{idx}"))
            self.amp_layers.append(AngularMaxPooling())

        #########
        # Output
        #########
        if segmentation_classes:
            self.output_dense = keras.layers.Dense(segmentation_classes, name="output")
        else:
            self.output_dense = keras.layers.Dense(6890, name="output")

        ###################
        # Training metrics
        ###################
        self.loss_tracker = keras.metrics.Mean(name="loss")
        self.accuracy = keras.metrics.SparseCategoricalAccuracy()

    def call(self, inputs, **kwargs):
        #################
        # Handling Input
        #################
        signal, bc = inputs
        signal = self.normalize(signal)
        signal = self.downsize_dense(signal)
        signal = self.downsize_bn(signal)

        ###############
        # Forward pass
        ###############
        for idx in range(len(self.output_dims)):
            signal = self.do_layers[idx](signal)
            signal = self.isc_layers[idx]([signal, bc])
            signal = self.amp_layers[idx](signal)
            signal = self.bn_layers[idx](signal)

        #########
        # Output
        #########
        return self.output_dense(signal)

    def test_step(self, data):
        (signal, bc), gt = data
        pred = self([signal, bc], training=False)
        loss = self.compute_loss(y=gt, y_pred=pred)

        # Statistics
        self.loss_tracker.update_state(loss)
        for metric in self.metrics[1:]:  # skip loss
            metric.update_state(gt, pred)
        return {m.name: m.result() for m in self.metrics}

    def train_step(self, data):
        (signal, bc), gt = data

        with tf.GradientTape() as tape:
            pred = self([signal, bc], training=True)
            loss = self.compute_loss(y=gt, y_pred=pred)

        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))

        # Statistics
        self.loss_tracker.update_state(loss)
        for metric in self.metrics[1:]:  # skip loss
            metric.update_state(gt, pred)

        return {m.name: m.result() for m in self.metrics}

    @property
    def metrics(self):
        # 'reset_states()' called automatically at the start of each training epoch or evaluation
        return [self.loss_tracker, self.accuracy]



from geoconv.preprocessing.barycentric_coordinates import polar_to_cart

from matplotlib import pyplot as plt
from matplotlib.collections import PolyCollection
from matplotlib.patches import Polygon
from PIL import Image

import os
import matplotlib
import trimesh
import numpy as np
import matplotlib.cm as cm
import io
import time


def draw_barycentric_coordinates(gpc_system, barycentric_coordinates, save_name=""):
    """Draws barycentric coordinates successively.

    Parameters
    ----------
    gpc_system: GPCSystem
        The GPC-system in which the barycentric coordinates should be illustrated.
    barycentric_coordinates: np.ndarray
        The barycentric coordinates to illustrate.
    save_name: str
        If given, the plot will be saved under this path.
    """
    interpolated_points = []
    for rc in range(barycentric_coordinates.shape[0]):
        for ac in range(barycentric_coordinates.shape[1]):
            triangle_indices = barycentric_coordinates[rc, ac, :, 0].astype(np.int16)
            triangle_interpolation_coefficients = barycentric_coordinates[rc, ac, :, 1]
            triangle_coordinates = np.stack(
                [gpc_system.x_coordinates[triangle_indices], gpc_system.y_coordinates[triangle_indices]],
                axis=-1
            )
            dot_interpolation = triangle_coordinates.T @ triangle_interpolation_coefficients
            interpolated_points.append(dot_interpolation)
            pc = ["blue" for _ in range(len(interpolated_points) - 1)] + ["yellow"] + ["orange" for _ in range(3)]
            draw_triangles(
                gpc_system.get_gpc_triangles(in_cart=True),
                points=np.array(interpolated_points + list(triangle_coordinates)),
                point_color=pc,
                title=f"Interpolation Coefficients: {triangle_interpolation_coefficients}",
                plot=True,
                save_name=f"{save_name}_{rc}_{ac}"
            )


def draw_multiple_princeton_benchmarks(save_name, **kwargs):
    """Draws the Princeton benchmarks all given numpy files.

    Parameters
    ----------
    save_name: str
        The name of the file in which the result plot will be saved.
    kwargs: str
        The keys will be displayed as a title. The values are the paths to the numpy files, the line-style
        and the color.
    """

    for name, (path, line_style, color) in kwargs.items():
        # Load values from princeton benchmark
        pb_values = np.load(path)

        # Filter values
        unique_x_values = np.unique(pb_values[:, 1])
        unique_values = []
        for unique_x in unique_x_values:
            unique_values.append(pb_values[np.where(pb_values[:, 1] == unique_x)[0][-1]])
        unique_values = np.array(unique_values)

        # Plot values
        plt.plot(unique_values[:, 1], unique_values[:, 0], linestyle=line_style, label=name, c=color)
        # plt.title(f"Princeton Benchmarks")
        plt.xlabel("geodesic error")
        plt.ylabel("\% correct correspondences")

    plt.grid()
    plt.legend()
    plt.savefig(f"{save_name}.svg")
    plt.show()


def draw_prior(icnn_layer, indices):
    """Wrapper method for 'draw_interpolation_coefficients_single_idx'

    Parameters
    ----------
    icnn_layer: src.layers.conv_intrinsic.ConvIntrinsic
        The for which the interpolation coefficients shall be visualized
    indices: List[int]
        A list of index-tuple for accessing template vertices. I.e. indices[x] = (a, b) with K[a, b] = (rho, theta).
    """
    fig = plt.figure(figsize=(10, 5))
    rows = len(indices)
    ax_idx = 0
    for r_idx, a_idx in indices:
        ax_idx += 1
        axis_ic = fig.add_subplot(rows, 2, ax_idx)
        ax_idx += 1
        axis_kv = fig.add_subplot(rows, 2, ax_idx, projection="polar")
        draw_interpolation_coefficients_single_idx(icnn_layer, r_idx, a_idx, fig, axis_ic, axis_kv)
    fig.tight_layout()
    plt.show()


def draw_interpolation_coefficients_single_idx(icnn_layer, radial_idx, angular_idx, fig, axis_ic, axis_kv):
    """Visualizes the interpolation coefficients of the patch operator at a specific template vertex

    Parameters
    ----------
    icnn_layer: src.layers.conv_intrinsic.ConvIntrinsic
        The for which the interpolation coefficients shall be visualized
    radial_idx: int
        The index of the radial coordinate from the template vertex for which we visualize the interpolation
        coefficients
    angular_idx: int
        The index of the angular coordinate from the template vertex for which we visualize the interpolation
        coefficients
    fig:
        The figure in which to plot the axes
    axis_ic:
        The axis on which to plot the interpolation coefficients matrix
    axis_kv:
        The axis on which to plot the weighted template vertices
    """

    # Get interpolation coefficients of the given layer: I[a, b] \in R^{n_radial * n_angular}
    weights = icnn_layer._kernel[radial_idx, angular_idx].numpy()
    template_size = icnn_layer._template_size
    template_matrix = icnn_layer._template_vertices.numpy()

    # Reshape vector into matrix: (n_radial * n_angular,) -> (n_radial, n_angular)
    # See 'ConvIntrinsic._configure_patch_operator()' for why it is stored as a vector.
    weights = weights.reshape(template_size)

    # Visualize interpolation coefficient matrix
    pos = axis_ic.matshow(weights, cmap="rainbow")
    fig.colorbar(pos, ax=axis_ic, fraction=0.046, pad=0.04)
    axis_ic.set_title(f"Prior as Matrix")
    template_matrix = template_matrix.reshape((-1, 2))

    # TODO: For some reason, matplotlib only shows all labels from list[1:] and forgets about list[0]
    axis_ic.set_ylabel("Radial coordinate")
    axis_ic.set_yticklabels(["0"] + [f"{x:.3f}" for x in np.unique(template_matrix[:, 0])])

    axis_ic.set_xlabel("Angular coordinate (in radian)")
    axis_ic.set_xticklabels(["0"] + [f"{x:.3f}" for x in np.unique(template_matrix[:, 1])])

    # Visualize interpolation coefficients at template vertices
    axis_kv.scatter(
        template_matrix[:, 1], template_matrix[:, 0], c=cm.rainbow(weights.flatten()), s=150, edgecolor="black"
    )
    axis_kv.set_title("Prior on Template Vertices")
    axis_kv.grid(True)
    axis_kv.set_axisbelow(True)


def draw_correspondences(query_mesh, prediction, reference_mesh, color_map="Reds", save_image=True):
    """Draw point correspondences between a query- and a reference mesh

    The point correspondence problem can be defined as labeling all vertices of a query
    mesh with the indices of the corresponding points in a reference mesh. See:

    > [Geodesic Convolutional Neural Networks on Riemannian Manifolds](https://arxiv.org/abs/1501.06297)
    > Jonathan Masci and Davide Boscaini et al.

    Parameters
    ----------
    query_mesh: trimesh.Trimesh
        The mesh that contains the vertices, which you want to label
    prediction: np.ndarray
        The predicted labels for the vertices in the query mesh
    reference_mesh: trimesh.Trimesh
        The reference mesh
    color_map: str
        The used color map. Checkout 'matplotlib' for available color maps.
    save_image: bool
        Whether to save the image
    """
    shift_dim = 0
    query_mesh.visual.vertex_colors = [100, 100, 100, 100]
    reference_mesh.visual.vertex_colors = [100, 100, 100, 100]

    ref_colors = trimesh.visual.interpolate(reference_mesh.vertices[:, shift_dim], color_map=color_map)
    reference_mesh_pc = trimesh.PointCloud(vertices=reference_mesh.vertices, colors=ref_colors)

    pred_colors = ref_colors[prediction]
    query_mesh.vertices[:, shift_dim] -= np.abs(
        query_mesh.vertices[:, shift_dim].min() - query_mesh.vertices[:, shift_dim].max()
    )
    query_mesh_pc = trimesh.PointCloud(vertices=query_mesh.vertices, colors=pred_colors)

    scene = trimesh.Scene([query_mesh, query_mesh_pc, reference_mesh, reference_mesh_pc])
    scene.show()
    if save_image:
        image_bytes = scene.save_image(resolution=(1080, 1080))
        image_array = np.array(Image.open(io.BytesIO(image_bytes)))
        matplotlib.image.imsave("./correspondence.png", image_array)


def draw_gpc_on_mesh(center_vertex,
                     radial_coordinates,
                     angular_coordinates,
                     object_mesh,
                     save_name="",
                     angles=(0., 0., 0.),
                     distance=1.,
                     center=(0., 0., 0.)):
    """Visualizes the radial and angular coordinates of a local GPC-system on an object mesh.

    This function first shows you the radial coordinates and then the angular coordinates.

    Parameters
    ----------
    center_vertex: int
        The index of the center vertex of the GPC-system.
    radial_coordinates: np.ndarray
        A 1-dimensional array containing the radial coordinates of the GPC-system.
    angular_coordinates: np.ndarray
        A 1-dimensional array containing the angular coordinates of the GPC-system.
    object_mesh: trimesh.Trimesh
        The object mesh.
    save_name: str
        The name of the image. If none is given, the image will not be saved.
    angles: tuple
        A 3-tuple that describes the camera angle in radians.
    distance: float
        A float that describes the camera distance.
    center: tuple
        A 3-tuple that describes the camera center.
    """
    radial_coordinates = radial_coordinates.copy()
    angular_coordinates = angular_coordinates.copy()

    object_mesh.visual.vertex_colors = [90, 90, 90, 200]

    # Visualize radial coordinates
    radial_coordinates[radial_coordinates == np.inf] = 0.0
    colors = trimesh.visual.interpolate(radial_coordinates, color_map="Reds")
    colors[center_vertex] = np.array([255, 255, 0, 255])
    point_cloud_1 = trimesh.points.PointCloud(object_mesh.vertices, colors=colors)
    to_visualize = [object_mesh, point_cloud_1]
    scene = trimesh.Scene(to_visualize)
    scene.set_camera(angles=angles, distance=distance, center=center)
    scene.show()
    if save_name:
        image_bytes = scene.save_image(resolution=(3840, 3840))
        image_array = np.array(Image.open(io.BytesIO(image_bytes)))
        matplotlib.image.imsave(f"{save_name}_radial_coords.png", image_array)

    # Visualize angular coordinates
    colors = trimesh.visual.interpolate(angular_coordinates, color_map="YlGn")
    colors[center_vertex] = np.array([255, 255, 0, 255])
    point_cloud_2 = trimesh.points.PointCloud(object_mesh.vertices, colors=colors)
    to_visualize.append(point_cloud_2)
    scene = trimesh.Scene(to_visualize)
    scene.set_camera(angles=angles, distance=distance, center=center)
    scene.show()
    if save_name:
        image_bytes = scene.save_image(resolution=(3840, 3840))
        image_array = np.array(Image.open(io.BytesIO(image_bytes)))
        matplotlib.image.imsave(f"{save_name}_angular_coords.png", image_array)


def draw_triangles(triangles, points=None, point_color="blue", title="", plot=True, save_name=""):
    """Draws a single triangle and optionally a point in 2D space.

    Parameters
    ----------
    triangles: np.ndarray
        The triangles in cartesian coordinates
    points: np.ndarray
        Points that can optionally also be visualized (in cartesian coordinates)
    point_color: str
        The point color
    title: str
        The title of the plot
    plot: bool
        Whether to plot the image immediately
    save_name: str
        If given, the plot will be stored under this path.
    """
    fig, ax = plt.subplots(1, 1)
    ax.set_title(title)
    for tri in triangles:
        polygon = Polygon(tri, alpha=.4, edgecolor="red")
        ax.add_patch(polygon)

    if points is not None:
        ax.scatter(points[:, 0], points[:, 1], color=point_color)

    if points is None:
        ax.set_xlim(triangles[:, :, 0].min(), triangles[:, :, 0].max())
        ax.set_ylim(triangles[:, :, 1].min(), triangles[:, :, 1].max())

    plt.grid()
    if save_name:
        plt.savefig(f"{save_name}.svg")

    if plot:
        plt.show()


def draw_gpc_triangles(gpc_system,
                       template_matrix=None,
                       alpha=.4,
                       edge_color="red",
                       scatter_color="green",
                       highlight_face=-1,
                       plot=True,
                       title="",
                       save_name=""):
    """Draws the triangles of a local GPC-system.

    Parameters
    ----------
    gpc_system: GPCSystem
        The GPC-system to visualize.
    template_matrix: np.ndarray
        A 3D-array that describes template vertices in cartesian coordinates. If 'None' is passed
        no template vertices will be visualized.
    alpha: float
        The opacity of the polygons
    edge_color: str
        The color for the triangle edges
    scatter_color: str
        The color for the template vertices (in case a template is given)
    highlight_face: int
        The index of a triangle, which shall be highlighted
    plot: bool
        Whether to immediately plot
    title: str
        The title of the plot
    save_name: str
        The name of the image. If none is given, the image will not be saved.
    """
    gpc_system_faces = gpc_system.get_gpc_triangles(in_cart=True)

    min_coordinate = gpc_system_faces.min()
    max_coordinate = gpc_system_faces.max()

    fig, ax = plt.subplots()
    ax.set_title(title)
    ax.set_xlim([min_coordinate, max_coordinate])
    ax.set_ylim([min_coordinate, max_coordinate])
    polygons = PolyCollection(gpc_system_faces, alpha=alpha, edgecolors=edge_color)
    ax.add_collection(polygons)

    # Print template
    if template_matrix is not None:
        for radial_idx in range(template_matrix.shape[0]):
            ax.scatter(template_matrix[radial_idx, :, 0], template_matrix[radial_idx, :, 1], color=scatter_color)

    # Highlight triangle
    if highlight_face > -1:
        ax.add_patch(
            Polygon(gpc_system_faces[highlight_face], linewidth=3., fill=False, edgecolor="purple")
        )
        ax.scatter(
            gpc_system_faces[highlight_face][:, 0],
            gpc_system_faces[highlight_face][:, 1],
            s=90.,
            color="purple"
        )
        for idx, annotation in enumerate(["a", "b", "c"]):
            x = gpc_system_faces[highlight_face][idx, 0]
            y = gpc_system_faces[highlight_face][idx, 1]
            ax.annotate(annotation, (x, y), fontsize=15)

    if save_name:
        plt.savefig(save_name)

    if plot:
        plt.show()

    plt.close(fig)
    return gpc_system


def draw_vertices_in_coordinate_system(radial_coordinates, angular_coordinates):
    """Plots the vertices of a GPC-system in a polar coordinate system.

    Parameters
    ----------
    radial_coordinates: np.ndarray
        A 1-dimensional array containing the radial coordinates of the GPC-system.
    angular_coordinates: np.ndarray
        A 1-dimensional array containing the angular coordinates of the GPC-system.
    """

    # Plot vertices
    mask = radial_coordinates != np.inf
    _, ax = plt.subplots(subplot_kw={'projection': 'polar'})
    ax.plot(angular_coordinates[mask], radial_coordinates[mask], "ro")
    plt.show()


def draw_edge_cache(edge_cache,
                    u,
                    theta,
                    edges_to_highlight=None,
                    point_to_highlight=None,
                    highlighting_color="red",
                    saving_folder="./visualization"):
    if not os.path.exists(saving_folder):
        os.makedirs(saving_folder)
    fig, ax = plt.subplots()
    for edge in edge_cache[-1]:
        vertex_1 = polar_to_cart(angles=theta[edge[0]], scales=u[edge[0]])
        vertex_2 = polar_to_cart(angles=theta[edge[1]], scales=u[edge[1]])
        ax.plot([vertex_1[0], vertex_2[0]], [vertex_1[1], vertex_2[1]], c="blue")
        ax.annotate(f"{edge[0]}", vertex_1)
        ax.annotate(f"{edge[1]}", vertex_2)
    if edges_to_highlight is not None:
        for edge in edges_to_highlight:
            vertex_1 = polar_to_cart(angles=theta[edge[0]], scales=u[edge[0]])
            vertex_2 = polar_to_cart(angles=theta[edge[1]], scales=u[edge[1]])
            ax.plot([vertex_1[0], vertex_2[0]], [vertex_1[1], vertex_2[1]], c=highlighting_color)
            ax.scatter([vertex_1[0], vertex_2[0]], [vertex_1[1], vertex_2[1]], c=highlighting_color)
            ax.annotate(f"{edge[0]}", vertex_1)
            ax.annotate(f"{edge[1]}", vertex_2)
    if point_to_highlight is not None:
        x, y = polar_to_cart(angles=point_to_highlight[2], scales=point_to_highlight[1])
        ax.scatter([x], [y], c="green")
        ax.annotate(point_to_highlight[0], [x, y])
    plt.savefig(f"./{saving_folder}/{time.time()}.svg")
    plt.close()
    time.sleep(.5)



from geoconv.tensorflow.layers.conv_intrinsic import ConvIntrinsic

import numpy as np


class ConvZero(ConvIntrinsic):
    """No interpolation weighting"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def define_kernel_values(self, template_matrix):
        """Returns all-zero interpolation coefficients, which causes the layer to only work with self-connections."""

        return np.zeros(template_matrix.shape[:-1] + template_matrix.shape[:-1])



from geoconv_examples.mpi_faust.tensorflow.faust_data_set import load_preprocessed_faust
from geoconv_examples.mpi_faust.tensorflow.model import Imcnn
from geoconv_examples.mpi_faust.data.preprocess_faust import preprocess_faust
from geoconv.utils.measures import princeton_benchmark

from pathlib import Path

import tensorflow as tf
import keras
import numpy as np


def train_model(reference_mesh_path,
                signal_dim,
                preprocessed_data,
                n_radial=5,
                n_angular=8,
                registration_path="",
                compute_shot=True,
                geodesic_diameters_path="",
                precomputed_gpc_radius=0.037,
                template_radius=0.027744965069279016,
                logging_dir="./imcnn_training_logs",
                processes=1,
                init_lr=0.00165,
                weight_decay=0.005,
                layer_conf=None,
                model="dirac",
                add_noise=False,
                reference_mesh_diameter=2.2093810817030244):
    """Trains one singular IMCNN

    Parameters
    ----------
    reference_mesh_path: str
        The path to the reference mesh file.
    signal_dim: int
        The dimensionality of the mesh signal
    preprocessed_data: str
        The path to the pre-processed data. If you have not pre-processed your data so far and saved it under the given
        path, this script will execute pre-processing for you. For this to work, you need to pass the arguments which
        are annotated with '[REQUIRED FOR PRE-PROCESSING]'. If pre-processing is not required, you can ignore those
        arguments.
    reference_mesh_diameter: float
        [REQUIRED FOR BENCHMARKING] The geodesic diameter of the reference mesh. Defaults to the diameter of the first
        registration mesh (tr_reg_000.ply) in the training set of the FAUST data set. If other reference mesh is
        selected, adjust this parameter accordingly!
    n_radial: int
        [REQUIRED FOR PRE-PROCESSING] The amount of radial coordinates for the template.
    n_angular: int
        [REQUIRED FOR PRE-PROCESSING] The amount of angular coordinates for the template.
    registration_path: str
        [REQUIRED FOR PRE-PROCESSING] The path of the training-registration files in the FAUST data set.
    compute_shot: bool
        [REQUIRED FOR PRE-PROCESSING] Whether to compute SHOT-descriptors during preprocessing as the mesh signal
    geodesic_diameters_path: str
        [REQUIRED FOR PRE-PROCESSING] The path to pre-computed geodesic diameters for the FAUST-registration meshes.
    precomputed_gpc_radius: float
        [REQUIRED FOR PRE-PROCESSING] The GPC-system radius to use for GPC-system computation. If not provided, the
        script will calculate it.
    template_radius: float
        [OPTIONAL] The template radius of the ISC-layer (the one used during preprocessing, defaults to radius for FAUST
        data set).
    logging_dir: str
        [OPTIONAL] The path to the folder where logs will be stored
    processes: int
        [OPTIONAL] The amount of concurrent processes. Affects preprocessing and Princeton benchmark.
    init_lr: float
        [OPTIONAL] Initial learning rate.
    weight_decay: float
        [OPTIONAL] Weight decay.
    layer_conf: list
        [OPTIONAL] List of tuples: The first entry references the output dimensions of the i-th ISC-layer, The second
        entry references of skips between each rotation while computing the convolution (rotation delta).
    model: str
        [OPTIONAL] Which model variant (['dirac', 'geodesic', 'zero']) shall be tuned.
    add_noise: bool
        [OPTIONAL] Adds Gaussian noise to the mesh data.
    """
    # Load data
    preprocess_zip = f"{preprocessed_data}.zip"
    if not Path(preprocess_zip).is_file():
        template_radius = preprocess_faust(
            n_radial=n_radial,
            n_angular=n_angular,
            target_dir=preprocess_zip[:-4],
            registration_path=registration_path,
            shot=compute_shot,
            geodesic_diameters_path=geodesic_diameters_path,
            precomputed_gpc_radius=precomputed_gpc_radius,
            processes=processes,
            add_noise=add_noise
        )
    else:
        print(f"Found preprocess-results: '{preprocess_zip}'. Skipping preprocessing.")

    seeds = [10, 20, 30, 40, 50]
    for exp_number in range(len(seeds)):
        # Set seeds
        tf.random.set_seed(seeds[exp_number])
        np.random.seed(seeds[exp_number])

        # Load data
        kernel_size = (n_radial, n_angular)
        train_data = load_preprocessed_faust(preprocess_zip, signal_dim=signal_dim, kernel_size=kernel_size, set_type=0)
        val_data = load_preprocessed_faust(preprocess_zip, signal_dim=signal_dim, kernel_size=kernel_size, set_type=1)

        # Define and compile model
        imcnn = Imcnn(
            signal_dim=signal_dim,
            kernel_size=kernel_size,
            template_radius=template_radius,
            layer_conf=layer_conf,
            variant=model
        )
        loss = keras.losses.SparseCategoricalCrossentropy(from_logits=True)
        opt = keras.optimizers.AdamW(
            learning_rate=keras.optimizers.schedules.ExponentialDecay(
                initial_learning_rate=init_lr,
                decay_steps=500,
                decay_rate=0.99
            ),
            weight_decay=weight_decay
        )
        imcnn.compile(optimizer=opt, loss=loss)

        # Adapt normalization
        print("Initializing normalization layer..")
        imcnn.normalize.build(tf.TensorShape([6890, signal_dim]))
        adaption_data = load_preprocessed_faust(
            preprocess_zip, signal_dim=signal_dim, kernel_size=kernel_size, set_type=0, only_signal=True
        )
        imcnn.normalize.adapt(adaption_data)
        print("Done.")

        # Build model
        imcnn([tf.random.uniform(shape=(6890, signal_dim)), tf.random.uniform(shape=(6890,) + kernel_size + (3, 2))])
        imcnn.summary()

        # Define callbacks
        csv = keras.callbacks.CSVLogger(f"{logging_dir}/training_{exp_number}.log")
        stop = keras.callbacks.EarlyStopping(monitor="val_loss", patience=20)
        tb = keras.callbacks.TensorBoard(
            log_dir=f"{logging_dir}/tensorboard_{exp_number}",
            histogram_freq=1,
            write_graph=False,
            write_steps_per_second=True,
            update_freq="epoch",
            profile_batch=(1, 70)
        )

        # Train and save model
        imcnn.fit(x=train_data, callbacks=[stop, tb, csv], validation_data=val_data, epochs=200)
        imcnn.save(f"{logging_dir}/saved_imcnn_{exp_number}")

        # Evaluate model with Princeton benchmark
        test_dataset = load_preprocessed_faust(
            preprocess_zip, signal_dim=signal_dim, kernel_size=kernel_size, set_type=2
        )
        princeton_benchmark(
            imcnn=imcnn,
            test_dataset=test_dataset,
            ref_mesh_path=reference_mesh_path,
            file_name=f"{logging_dir}/model_benchmark_{exp_number}",
            processes=processes,
            geodesic_diameter=reference_mesh_diameter
        )



from geoconv.pytorch.layers.conv_geodesic import angle_distance
from geoconv.pytorch.layers.conv_intrinsic import ConvIntrinsic

import numpy as np
import scipy as sp


def exp_pdf(mean_rho, mean_theta, rho, theta, exp_lambda):
    """Exponential probability distribution for geodesic polar coordinates

    Parameters
    ----------
    mean_rho: float
        Mean radial distance of the normal
    mean_theta: float
        Mean angle for the Gaussian
    rho: float
        Radial coordinate of the interpolation point that shall be weighted
    theta: float
        Angular coordinate of the interpolation point that shall be weighted
    exp_lambda: float
        The lambda parameter for the exponential probability density function

    Returns
    -------
    float:
        The weight for the interpolation point (rho, theta)
    """

    # Compute delta theta
    max_angle = np.maximum(mean_theta, theta)
    min_angle = np.minimum(mean_theta, theta)
    delta_angle = angle_distance(max_angle, min_angle)

    # Compute delta rho
    delta_rho = np.abs(rho - mean_rho)

    return exp_lambda ** 2 * np.exp(-exp_lambda * (delta_rho + delta_angle))


class ConvExp(ConvIntrinsic):
    """Exponential vertex weighting"""
    def __init__(self, *args, exp_lambda=1, **kwargs):
        self.exp_lambda = exp_lambda
        super().__init__(*args, **kwargs)

    def define_kernel_values(self, template_matrix):
        template_matrix[:, :, 0] = template_matrix[:, :, 0] / template_matrix[:, :, 0].max()
        interpolation_coefficients = np.zeros(template_matrix.shape[:-1] + template_matrix.shape[:-1])
        for mean_rho_idx in range(template_matrix.shape[0]):
            for mean_theta_idx in range(template_matrix.shape[1]):
                mean_rho, mean_theta = template_matrix[mean_rho_idx, mean_theta_idx]
                for rho_idx in range(template_matrix.shape[0]):
                    for theta_idx in range(template_matrix.shape[1]):
                        rho, theta = template_matrix[rho_idx, theta_idx]
                        interpolation_coefficients[mean_rho_idx, mean_theta_idx, rho_idx, theta_idx] = exp_pdf(
                            mean_rho, mean_theta, rho, theta, self.exp_lambda
                        )
                interpolation_coefficients[mean_rho_idx, mean_theta_idx] = sp.special.softmax(
                    interpolation_coefficients[mean_rho_idx, mean_theta_idx]
                )
        return interpolation_coefficients



from geoconv_examples.mpi_faust.data.preprocess_faust import get_file_number

import numpy as np
import tensorflow as tf
import os
import random


def faust_generator(path_to_zip, set_type=0, only_signal=False, return_coordinates=False, set_indices=None):
    """Reads one element of preprocessed FAUST-geoconv_examples into memory per 'next'-call.

    Parameters
    ----------
    path_to_zip: str
        The path to the .zip-file that contains the preprocessed faust data set
    set_type: int
        This integer has to be either:
            - 0 -> "train"
            - 1 -> "validation"
            - 2 -> "test"
            - 3 -> "all"
        Depending on the choice, the training-, validation or testing data set will be returned. The split is equal to
        the one given in:
        > [Geodesic Convolutional Neural Networks on Riemannian Manifolds](https://arxiv.org/abs/1501.06297)
        > Jonathan Masci and Davide Boscaini et al.
    only_signal: bool
        Return only the signal matrices. Helpful for keras.Normalization(axis=-1).adapt(data)
    return_coordinates: bool
        Whether to return the coordinates of the mesh vertices. Requires coordinates to be contained in preprocessed
        dataset.
    set_indices: list
        A list of integer values that determine which meshes shall be returned. If it is set to 'None', the set
        type determine which meshes will be returned. Defaults to 'None'. Adds noise to barycentric coordinates
        if set type is set to 0.

    Returns
    -------
    generator:
        A generator yielding the preprocessed data. I.e. the signal defined on the vertices, the barycentric coordinates
        and the ground truth correspondences.
    """
    kernel_size = None
    dataset = np.load(path_to_zip, allow_pickle=True)
    file_names = [os.path.basename(fn) for fn in dataset.files]
    SIGNAL = [file_name for file_name in file_names if file_name.startswith("SIGNAL")]
    BC = [file_name for file_name in file_names if file_name.startswith("BC")]
    GT = [file_name for file_name in file_names if file_name.startswith("GT")]
    SIGNAL.sort(key=get_file_number), BC.sort(key=get_file_number), GT.sort(key=get_file_number)
    if return_coordinates:
        COORD = [file_name for file_name in file_names if file_name.startswith("COORD")]
        COORD.sort(key=get_file_number)

    # Set iteration indices according to set type
    if set_indices is None:
        if set_type == 0:
            indices = list(range(70))
            random.shuffle(indices)
        elif set_type == 1:
            indices = range(70, 80)
        elif set_type == 2:
            indices = range(80, 100)
        elif set_type == 3:
            indices = range(100)
        else:
            raise RuntimeError(
                f"There is no 'set_type'={set_type}. Choose from: [0: 'train', 1: 'val', 2: 'test', 3: 'all']."
            )
    else:
        indices = set_indices

    for idx in indices:
        # Read signal
        signal = tf.cast(dataset[SIGNAL[idx]], tf.float32)

        # Read bc + add noise
        bc = tf.cast(dataset[BC[idx]], tf.float32)
        if kernel_size is None:
            kernel_size = bc.shape[1:3]

        if set_type == 0:
            noise = np.abs(np.random.normal(size=(6890,) + kernel_size + (3, 2), scale=1e-5))
            noise[:, :, :, :, 0] = 0
            bc = bc + noise

        # Ground truth: Return the indices of the ones for each row
        # (as required by `keras.losses.SparseCategoricalCrossentropy`)
        gt = dataset[GT[idx]]

        if only_signal:
            yield signal
        else:
            if return_coordinates:
                coord = tf.cast(dataset[COORD[idx]], tf.float32)
                yield (signal, bc, coord), gt
            else:
                yield (signal, bc), gt


def load_preprocessed_faust(path_to_zip,
                            signal_dim,
                            kernel_size=(2, 4),
                            set_type=0,
                            only_signal=False,
                            return_coordinates=False):
    """Returns a 'tensorflow.data.Dataset' of the preprocessed MPI-FAUST geoconv_examples.

    Requires that preprocessing already happened. This function operates directly on the resulting 'zip'-file.

    Parameters
    ----------
    path_to_zip: str
        The path to the 'zip'-file obtained through preprocessing
    signal_dim: int
        The dimensionality of the defined signal
    kernel_size: tuple
        The same amount of radial- and angular coordinates of the kernel defined in the preprocessing step
    set_type: int
        This integer has to be either:
            - 0 -> "train"
            - 1 -> "validation"
            - 2 -> "test"
        Depending on the choice, the training-, validation or testing data set will be returned. The split is equal to
        the one given in:
        > [Geodesic Convolutional Neural Networks on Riemannian Manifolds](https://arxiv.org/abs/1501.06297)
        > Jonathan Masci and Davide Boscaini et al.
    only_signal: bool
        Return only the signal matrices. Helpful for keras.Normalization(axis=-1).adapt(data)
    return_coordinates: bool
        Whether to return the coordinates of the mesh vertices. Requires coordinates to be contained in preprocessed
        dataset.

    Returns
    -------
    tensorflow.data.Dataset:
        A tensorflow data set of the preprocessed MPI-FAUST geoconv_examples
    """
    if only_signal:
        output_signature = tf.TensorSpec(shape=(None, signal_dim,), dtype=tf.float32)
    else:
        if return_coordinates:
            output_signature = (
                (
                    tf.TensorSpec(shape=(None, signal_dim,), dtype=tf.float32),  # Signal
                    tf.TensorSpec(shape=(None,) + kernel_size + (3, 2), dtype=tf.float32),  # Barycentric Coordinates
                    tf.TensorSpec(shape=(None, 3,), dtype=tf.float32),  # Coordinates
                ),
                tf.TensorSpec(shape=(None,), dtype=tf.float32)
            )
        else:
            output_signature = (
                (
                    tf.TensorSpec(shape=(None, signal_dim,), dtype=tf.float32),  # Signal
                    tf.TensorSpec(shape=(None,) + kernel_size + (3, 2), dtype=tf.float32)  # Barycentric Coordinates
                ),
                tf.TensorSpec(shape=(None,), dtype=tf.float32)
            )
    return tf.data.Dataset.from_generator(
        faust_generator,
        args=(path_to_zip, set_type, only_signal, return_coordinates),
        output_signature=output_signature
    )



from distutils.core import setup, Extension

import numpy as np


c_extension = Extension(
    "c_extension",
    ["./src/geoconv/preprocessing/c_extension/c_extension.c"],
    include_dirs=[np.get_include()],
    extra_link_args=["-lblas", "-lcblas"]
)

if __name__ == "__main__":
    setup(
        ext_modules=[c_extension]
    )



from geoconv.preprocessing.barycentric_coordinates import compute_barycentric_coordinates
from geoconv.preprocessing.gpc_system_group import GPCSystemGroup
from geoconv.utils.misc import shuffle_mesh_vertices, normalize_mesh, find_largest_one_hop_dist

from pathlib import Path

import pyshot
import os
import trimesh
import numpy as np
import shutil
import json


def get_file_number(file_name):
    """Extracts the file number contained in the file name

    Parameters
    ----------
    file_name: str
        The file name

    Returns
    -------
    int:
        The file number contained in the file name
    """
    # file_name.split(".")[0] -> Without file ending
    for elem in file_name.split(".")[0].split("_"):
        if elem.isdigit():
            return int(elem)
    raise RuntimeError(f"Filename '{file_name}' has no digit.")


def preprocess_faust(n_radial,
                     n_angular,
                     target_dir,
                     registration_path,
                     shot=True,
                     geodesic_diameters_path="",
                     precomputed_gpc_radius=-1.,
                     processes=1,
                     add_noise=False):
    """Preprocesses the FAUST-data set

    The FAUST-data set has to be downloaded from: https://faust-leaderboard.is.tuebingen.mpg.de/

    It was published in:
    > [FAUST: Dataset and evaluation for 3D mesh registration.]
    (https://www.cv-foundation.org/openaccess/content_cvpr_2014/html/Bogo_FAUST_Dataset_and_2014_CVPR_paper.html)
    > Bogo, Federica, et al.

    Parameters
    ----------
    n_radial: int
        The amount of radial coordinates for the kernel in your geodesic convolution.
    n_angular: int
        The amount of angular coordinates for the kernel in your geodesic convolution.
    target_dir: str
        The path to the directory in which the preprocessing results will be stored. At the end of the preprocessing
        procedure this directory will be replaced with a zip-compressed archive.
    registration_path: str
        The path to the training-registration meshes of the FAUST-data set.
    shot: bool
        Whether to compute SHOT-descriptors as mesh signal.
    geodesic_diameters_path: str
        Path, which points to *.npy file, that contains the geodesic diameters for all the meshes in the dataset.
    precomputed_gpc_radius: float
        The GPC-system radius to use for GPC-system computation. If not provided, the script will calculate it.
    processes: int
        The amount of concurrent processes that compute GPC-systems.
    add_noise: bool
        Adds Gaussian noise to the mesh data.

    Returns
    -------
    float:
        The used kernel radius.
    """
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    ##################
    # Load mesh-paths
    ##################
    paths_reg_meshes = os.listdir(registration_path)
    paths_reg_meshes.sort(key=get_file_number)
    paths_reg_meshes = [f for f in paths_reg_meshes if f[-4:] != ".png"]

    # Create temp dir for normalized meshes
    temp_dir = "./temp_meshes"
    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)

    # Check whether geodesic diameters have already been computed
    if Path(geodesic_diameters_path).is_file():
        geodesic_diameters = np.load(geodesic_diameters_path)
        print(f"\n\nLoaded geodesic diameters from: '{geodesic_diameters_path}'\n\n")
    else:
        geodesic_diameters = np.full(len(paths_reg_meshes), -1.)

    # Determine GPC-system-radius
    gpc_radius = .0
    for file_idx in range(len(paths_reg_meshes)):
        # Define file names for normalized vertices and faces (=temp.-meshes)
        reg_file_name = f"{registration_path}/{paths_reg_meshes[file_idx]}"
        normalized_v_name = f"{temp_dir}/vertices_{file_idx}.npy"
        normalized_f_name = f"{temp_dir}/faces_{file_idx}.npy"
        reg_mesh = trimesh.load_mesh(reg_file_name)

        # Check whether normalized meshes already exist
        if not (Path(normalized_v_name).is_file() and Path(normalized_f_name).is_file()):
            # Center and normalize mesh to unit geodesic diameter
            if geodesic_diameters[file_idx] == -1.:
                reg_mesh, geodesic_diameter = normalize_mesh(reg_mesh)
                geodesic_diameters[file_idx] = geodesic_diameter
            else:
                reg_mesh, geodesic_diameter = normalize_mesh(reg_mesh, geodesic_diameters[file_idx])

            # Add noise
            if add_noise:
                reg_mesh.vertices = reg_mesh.vertices + np.random.normal(size=(6890, 3), loc=0, scale=0.0005)

            # Save normalized mesh
            np.save(normalized_v_name, np.asarray(reg_mesh.vertices))
            np.save(normalized_f_name, np.asarray(reg_mesh.faces))
        else:
            vertices = np.load(normalized_v_name)
            faces = np.load(normalized_f_name)
            reg_mesh = trimesh.Trimesh(vertices=vertices, faces=faces)
            print(f"Found temp-files:\n{normalized_v_name}\n{normalized_f_name}\nSkipping to next normalization..")

        if precomputed_gpc_radius == -1.:
            new_candidate = find_largest_one_hop_dist(reg_mesh)
            gpc_radius = new_candidate if new_candidate > gpc_radius else gpc_radius
    if precomputed_gpc_radius > 0:
        gpc_radius = precomputed_gpc_radius
    else:
        gpc_radius = gpc_radius + gpc_radius * .1
    kernel_radius = gpc_radius * 0.75
    print(f"GPC-system radius: {gpc_radius:.3f} | Kernel radius: {kernel_radius:.3f}")

    # Save computed geodesic diameters
    if not Path(geodesic_diameters_path).is_file():
        geodesic_diameters_name = f"{target_dir}/geodesic_diameters.npy"
        np.save(geodesic_diameters_name, geodesic_diameters)

    # Log GPC-system radius and kernel radius
    properties_file = open(f"{target_dir}/gpc_kernel_properties.json", "w")
    json.dump({"gpc_system_radius": gpc_radius, "kernel_radius": kernel_radius}, properties_file, indent=4)
    properties_file.close()

    for file_idx in range(len(paths_reg_meshes)):
        # Define file names
        bc_name = f"{target_dir}/BC_{paths_reg_meshes[file_idx][:-4]}.npy"
        gt_name = f"{target_dir}/GT_{paths_reg_meshes[file_idx][:-4]}.npy"
        signal_name = f"{target_dir}/SIGNAL_{paths_reg_meshes[file_idx][:-4]}.npy"

        # Load normalized mesh
        vertices = np.load(f"{temp_dir}/vertices_{file_idx}.npy")
        faces = np.load(f"{temp_dir}/faces_{file_idx}.npy")
        reg_mesh = trimesh.Trimesh(vertices=vertices, faces=faces)

        # Check whether preprocessed files already exist
        if not (Path(bc_name).is_file() and Path(gt_name).is_file() and Path(signal_name).is_file()):
            #######################################################
            # Shuffle vertices of query mesh and save ground truth
            #######################################################
            reg_mesh, _, ground_truth = shuffle_mesh_vertices(reg_mesh)
            np.save(gt_name, ground_truth)

            ####################
            # Store mesh signal
            ####################
            if shot:
                radius = find_largest_one_hop_dist(reg_mesh) * 2.5
                shot_descrs = pyshot.get_descriptors(
                    reg_mesh.vertices,
                    reg_mesh.faces,
                    radius=radius,
                    local_rf_radius=radius,
                    min_neighbors=10,
                    n_bins=16,
                    double_volumes_sectors=True,
                    use_interpolation=True,
                    use_normalization=True
                )
                np.save(signal_name, shot_descrs)
            else:
                np.save(signal_name, np.asarray(reg_mesh.vertices))

            ############################
            # Compute local GPC-systems
            ############################
            gpc_systems = GPCSystemGroup(reg_mesh, processes=processes)
            gpc_systems.compute(u_max=gpc_radius)

            ##################################
            # Compute Barycentric coordinates
            ##################################
            bary_coords = compute_barycentric_coordinates(
                gpc_systems, n_radial=n_radial, n_angular=n_angular, radius=kernel_radius
            )
            np.save(bc_name, bary_coords)
        else:
            print(f"Found temp-files:\n{bc_name}\n{gt_name}\n{signal_name}\nSkipping to next temp.-mesh..")

    shutil.rmtree(temp_dir)
    shutil.make_archive(target_dir, "zip", target_dir)
    shutil.rmtree(target_dir)
    print("Preprocessing finished.")

    return kernel_radius



from geoconv.utils.misc import get_faces_of_edge, compute_vector_angle

from scipy.linalg import blas

import c_extension
import numpy as np


def compute_u_ijk_and_angle(vertex_i, vertex_j, vertex_k, u, theta, object_mesh, use_c, rotation_axis):
    """Euclidean update procedure for a vertex i in a given triangle and angle computation

    See Section 3 in:

    > [Geodesic polar coordinates on polygonal meshes]
      (https://onlinelibrary.wiley.com/doi/full/10.1111/j.1467-8659.2012.03187.x)
    > Melvær, Eivind Lyche, and Martin Reimers.

    Parameters
    ----------
    vertex_i: int
        The index of the vertex for which we want to update the distance and angle
    vertex_j: int
        The index of the second vertex in the triangle of vertex i
    vertex_k: int
        The index of the third vertex in the triangle of vertex i
    u: np.ndarray
        The currently known radial coordinates
    theta: np.ndarray
        The currently known angular coordinates
    object_mesh: trimesh.Trimesh
        A loaded object mesh
    use_c: bool
        A flag whether to use the c-extension
    rotation_axis: np.ndarray [DEPRECATED]
        The vertex normal of the center vertex from the considered GPC-system

    Returns
    -------
    (float, float)
        The Euclidean update u_ijk for vertex i (see equation 13 in paper) and the new angle vertex i
    """

    # Convert indices to vectors
    u_j, u_k = u[[vertex_j, vertex_k]]
    theta_i_init, theta_j, theta_k = theta[[vertex_i, vertex_j, vertex_k]]
    vertex_i, vertex_j, vertex_k = object_mesh.vertices[[vertex_i, vertex_j, vertex_k]]

    if use_c:
        result = np.array([0., 0.])
        c_extension.compute_dist_and_dir(
            result,
            vertex_i,
            vertex_j,
            vertex_k,
            u_j,
            u_k,
            theta_j,
            theta_k,
            rotation_axis
        )
        u_ijk, theta_i = result
    else:
        e_j = np.empty(3)
        blas.dcopy(vertex_j, e_j)
        blas.daxpy(vertex_i, e_j, a=-1.0)
        e_j_norm = blas.dnrm2(e_j)

        e_k = np.empty(3)
        blas.dcopy(vertex_k, e_k)
        blas.daxpy(vertex_i, e_k, a=-1.0)
        e_k_norm = blas.dnrm2(e_k)

        e_kj = np.empty(3)
        blas.dcopy(vertex_k, e_kj)
        blas.daxpy(vertex_j, e_kj, a=-1.0)
        e_kj_sqnrm = blas.ddot(e_kj, e_kj)

        A = e_j_norm * e_k_norm * np.sin(compute_vector_angle(e_j, e_k, None))

        # variant 1:
        square_1, square_2 = np.square(np.array([u_j - u_k, u_j + u_k]))
        radicand = (e_kj_sqnrm - square_1) * (square_2 - e_kj_sqnrm)

        # variant 2:
        # c, b, a = np.sort(np.array([u_j, u_k, blas.dnrm2(e_kj)]))
        # radicand = (a + (b + c)) * (c - (a - b)) * (c + (a - b)) * (a + (b - c))

        if radicand <= 0:
            j = u_j + blas.dnrm2(e_j)
            k = u_k + blas.dnrm2(e_k)
            if j <= k:
                u_ijk = j
                theta_i = theta_j
            else:
                u_ijk = k
                theta_i = theta_k
        else:
            H = np.sqrt(radicand)
            u_j_sq, u_k_sq = np.square(np.array([u_j, u_k]))
            x_j = A * (e_kj_sqnrm + u_k_sq - u_j_sq) + blas.ddot(e_k, e_kj) * H
            x_k = A * (e_kj_sqnrm + u_j_sq - u_k_sq) - blas.ddot(e_j, e_kj) * H
            # If x_k < 0 or x_k < 0 then alpha > 1, causing theta_i to be negative (and we don't want that).
            if x_j < 0 or x_k < 0:
                j = u_j + blas.dnrm2(e_j)
                k = u_k + blas.dnrm2(e_k)
                if j <= k:
                    u_ijk = j
                    theta_i = theta_j
                else:
                    u_ijk = k
                    theta_i = theta_k
            else:
                # Compute distance
                denominator = 2 * A * e_kj_sqnrm
                x_j, x_k = np.array([x_j, x_k]) / denominator
                blas.dscal(x_j, e_j)
                blas.dscal(x_k, e_k)

                result_vector = np.empty(3)
                blas.dcopy(e_k, result_vector)
                blas.daxpy(e_j, result_vector)
                u_ijk = blas.dnrm2(result_vector)

                # Compute angle
                s = np.empty(3)
                blas.dcopy(result_vector, s)
                blas.daxpy(vertex_i, s)

                blas.daxpy(s, vertex_k, a=-1.0)
                blas.daxpy(s, vertex_j, a=-1.0)
                blas.daxpy(s, vertex_i, a=-1.0)

                phi_kj = compute_vector_angle(vertex_k, vertex_j, None)
                phi_ij = compute_vector_angle(vertex_i, vertex_j, None)
                alpha = phi_ij / phi_kj

                # Pay attention to 0-2pi-discontinuity
                if theta_k <= theta_j:
                    if theta_j - theta_k >= np.pi:
                        theta_k = theta_k + 2 * np.pi
                else:
                    if theta_k - theta_j >= np.pi:
                        theta_j = theta_j + 2 * np.pi
                theta_i = np.fmod((1 - alpha) * theta_j + alpha * theta_k, 2 * np.pi)

    return u_ijk, theta_i


def compute_distance_and_angle(vertex_i, vertex_j, gpc_system, use_c, rotation_axis):
    """Euclidean update procedure for geodesic distance approximation

    See Section 4 in:

    > [Geodesic polar coordinates on polygonal
    meshes](https://onlinelibrary.wiley.com/doi/full/10.1111/j.1467-8659.2012.03187.x)
    > Melvær, Eivind Lyche, and Martin Reimers.

    Parameters
    ----------
    vertex_i: int
        The index of the vertex for which we want to update the distance and angle
    vertex_j: int
        The index of a second vertex in the triangle of vertex i (candidate vertex from heap)
    gpc_system: GPCSystem
        The current GPC-system.
    use_c: bool
        A flag whether to use the c-extension
    rotation_axis: np.ndarray [DEPRECATED]
        The vertex normal of the center vertex from the considered GPC-system

    Returns
    -------
    (float, float, list)
        The Euclidean update u_ijk for vertex i (see equation 13 in paper) and the new angle vertex i. Lastly, this
        function also returns the missing vertices which could have been used to update the coordinates of `vertex_i`.
    """
    # We consider both faces of `sorted_edge` for computing the coordinates to `vertex_i`
    sorted_edge = np.sort([vertex_i, vertex_j])
    if (sorted_edge[0], sorted_edge[1]) in gpc_system.faces.keys():
        # Use cache to get faces of `sorted_edge`
        considered_faces = gpc_system.faces[(sorted_edge[0], sorted_edge[1])]
    else:
        _, considered_faces = get_faces_of_edge(sorted_edge, gpc_system.object_mesh)

    # Compute GPC for `vertex_i` considering both faces of `[vertex_i, vertex_j]`
    updates = []
    k_vertices = []
    for face in considered_faces:
        vertex_k = [v for v in face if v not in [vertex_i, vertex_j]][0]
        k_vertices.append(vertex_k)
        # We need to know the distance to `vertex_k`
        if gpc_system.radial_coordinates[vertex_k] < np.inf and gpc_system.angular_coordinates[vertex_k] >= 0.:
            u_ijk, phi_i = compute_u_ijk_and_angle(
                vertex_i,
                vertex_j,
                vertex_k,
                gpc_system.radial_coordinates,
                gpc_system.angular_coordinates,
                gpc_system.object_mesh,
                use_c,
                rotation_axis
            )
            updates.append((u_ijk, phi_i, vertex_k))

    # If no GPC have been found for `vertex_i`, return default GPC
    if not updates:
        return np.inf, -1.0, None

    # If two GPC have been found for `vertex_i`, return the smallest distance to `vertex_i`
    else:
        u_ijk, phi_i, vertex_k = min(updates)
        return u_ijk, phi_i, k_vertices



from geoconv.pytorch.layers.conv_geodesic import angle_distance
from geoconv.pytorch.layers.conv_intrinsic import ConvIntrinsic

import numpy as np
import scipy as sp


def gamma_func(dof):
    """Computes the gamma value for the given degrees of freedom

    Parameters
    ----------
    dof: int
        Degrees of freedom for chi-squared distribution

    Returns
    -------
    float:
        The gamma value for the given degrees of freedom
    """
    assert dof >= 1, "You need to have at least one degree of freedom."
    r = dof / 2
    if dof == 1/2:
        return np.sqrt(np.pi)
    elif dof == 1:
        return 1.
    else:
        return r * gamma_func(dof - 1)


def chi_squared_pdf(mean_rho, mean_theta, rho, theta, dof):
    """Exponential probability distribution for geodesic polar coordinates

    Parameters
    ----------
    mean_rho: float
        Mean radial distance of the normal
    mean_theta: float
        Mean angle for the Gaussian
    rho: float
        Radial coordinate of the interpolation point that shall be weighted
    theta: float
        Angular coordinate of the interpolation point that shall be weighted
    dof: int
        Degrees of freedom for chi-squared distribution

    Returns
    -------
    float:
        The weight for the interpolation point (rho, theta)
    """
    assert dof >= 1, "You need to have at least one degree of freedom."

    # Compute delta theta
    max_angle = np.maximum(mean_theta, theta)
    min_angle = np.minimum(mean_theta, theta)
    delta_angle = angle_distance(max_angle, min_angle)
    if delta_angle == 0 and dof == 1:
        return 1.
    delta_angle_p = delta_angle ** (dof / 2 - 1)

    # Compute delta rho
    delta_rho = np.abs(rho - mean_rho)
    if delta_rho == 0 and dof == 1:
        return 1.
    delta_rho_p = delta_rho ** (dof / 2 - 1)

    gamma = (1 / (2 ** (dof / 2) * gamma_func(dof))) ** 2

    return gamma * delta_rho_p * delta_angle_p * np.exp(-(delta_rho + delta_angle) / 2)


class ConvChiSquared(ConvIntrinsic):
    """Chi-squared vertex weighting"""
    def __init__(self, *args, dof=2, **kwargs):
        self.dof = dof
        super().__init__(*args, **kwargs)

    def define_kernel_values(self, template_matrix):
        template_matrix[:, :, 0] = template_matrix[:, :, 0] / template_matrix[:, :, 0].max()
        interpolation_coefficients = np.zeros(template_matrix.shape[:-1] + template_matrix.shape[:-1])
        for mean_rho_idx in range(template_matrix.shape[0]):
            for mean_theta_idx in range(template_matrix.shape[1]):
                mean_rho, mean_theta = template_matrix[mean_rho_idx, mean_theta_idx]
                for rho_idx in range(template_matrix.shape[0]):
                    for theta_idx in range(template_matrix.shape[1]):
                        rho, theta = template_matrix[rho_idx, theta_idx]
                        interpolation_coefficients[mean_rho_idx, mean_theta_idx, rho_idx, theta_idx] = chi_squared_pdf(
                            mean_rho, mean_theta, rho, theta, self.dof
                        )
                interpolation_coefficients[mean_rho_idx, mean_theta_idx] = sp.special.softmax(
                    interpolation_coefficients[mean_rho_idx, mean_theta_idx]
                )
        return interpolation_coefficients
